#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
from multiprocessing import Pool
from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
from picai_eval import evaluate_folder
from picai_prep.preprocessing import crop_or_pad
from report_guided_annotation import extract_lesion_candidates

from picai_baseline.nnunet.softmax_export import \
    convert_cropped_npz_to_original_nifty
from picai_baseline.splits.picai import valid_splits as picai_pub_valid_splits
from picai_baseline.splits.picai_nnunet import \
    valid_splits as picai_pub_nnunet_valid_splits
from picai_baseline.splits.picai_pubpriv import \
    valid_splits as picai_pubpriv_valid_splits
from picai_baseline.splits.picai_pubpriv_nnunet import \
    valid_splits as picai_pubpriv_nnunet_valid_splits

try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass


def extract_lesion_candidates_cropped(pred, threshold):
    size = pred.shape
    pred = crop_or_pad(pred, (20, 384, 384))
    pred = crop_or_pad(pred, size)
    return extract_lesion_candidates(pred, threshold=threshold)[0]


def evaluate(
    task: str = "Task2201_picai_baseline",
    trainer: str = "nnUNetTrainerV2_Loss_FL_and_CE_checkpoints",
    workdir: Union[Path, str] = "/workdir",
    task_dir: Union[Path, str] = "auto",
    checkpoints: str = ["model_best"],
    folds: str = list(range(5)),
    softmax_postprocessing_func: "Optional[Union[Callable[[npt.NDArray[np.float_]], npt.NDArray[np.float_]], str]]" = "extract_lesion_candidates",
    threshold: str = "dynamic",
    metrics_fn: str = r"metrics-{checkpoint}-{threshold}.json",
    splits: str = "picai_pub",
    predictions_folder: str = r"picai_pubtrain_predictions_{checkpoint}",
    labels_folder: Union[Path, str] = "auto",
    verbose: int = 2,
):
    # input validation
    workdir = Path(workdir)
    if task_dir == "auto":
        task_dir = workdir / "results" / "nnUNet" / "3d_fullres" / task
    else:
        task_dir = workdir / task_dir

    if isinstance(splits, str):
        if splits == "":
            splits = None
        else:
            # select splits
            predefined_splits = {
                "picai_pub": picai_pub_valid_splits,
                "picai_pubpriv": picai_pubpriv_valid_splits,
                "picai_pub_nnunet": picai_pub_nnunet_valid_splits,
                "picai_pubpriv_nnunet": picai_pubpriv_nnunet_valid_splits,
            }
            if splits in predefined_splits:
                splits = predefined_splits[splits]
            else:
                # `splits` should be the path to a json file containing the splits
                print(f"Loading splits from {splits}")

    if isinstance(softmax_postprocessing_func, str):
        if softmax_postprocessing_func == "extract_lesion_candidates":
            def softmax_postprocessing_func(pred):
                return extract_lesion_candidates(pred, threshold=threshold)[0]
        elif softmax_postprocessing_func == "extract_lesion_candidates_cropped":
            def softmax_postprocessing_func(pred):
                return extract_lesion_candidates_cropped(pred, threshold=threshold)
        else:
            raise ValueError(f"Unrecognised softmax_postprocessing_func: {softmax_postprocessing_func}")

    for fold in folds:
        print(f"Evaluating fold {fold}...")

        for checkpoint in checkpoints:
            pred_folder = predictions_folder.replace(r"{checkpoint}", checkpoint)
            softmax_dir = task_dir / f"{trainer}__nnUNetPlansv2.1" / f"fold_{fold}" / pred_folder
            metrics_path = softmax_dir.parent / metrics_fn.replace(r"{checkpoint}", checkpoint).replace(r"{threshold}", threshold).replace("{fold}", str(fold))

            if metrics_path.exists():
                print(f"Metrics found at {metrics_path}, skipping..")
                continue
            else:
                print(f"Metrics will be saved to {metrics_path}.")

            original_softmax_prediction_paths = softmax_dir.glob("*.npz")

            if verbose >= 2:
                print(f"Predictions folder: {softmax_dir}")
                original_softmax_prediction_paths = list(original_softmax_prediction_paths)
                print(f"Found {len(original_softmax_prediction_paths)} predictions (e.g., {original_softmax_prediction_paths[0:2]})")

            # pad raw npz predictions to their original extent
            print("Converting nnU-Net predictions to original extent...")
            with Pool() as pool:
                pool.map(
                    func=convert_cropped_npz_to_original_nifty,
                    iterable=original_softmax_prediction_paths,
                )

            # select subject list
            if splits is None:
                subject_list = None
            elif isinstance(splits, dict):
                subject_list = splits[fold]['subject_list']
            elif isinstance(splits, str):
                path = Path(splits.replace(r"{fold}", str(fold)))
                with open(path, "r") as f:
                    splits = json.load(f)
                subject_list = splits['subject_list']
            else:
                raise ValueError(f"Unrecognised splits: {splits}")

            # select labels folder
            if labels_folder == "auto":
                y_true_dir = workdir / "nnUNet_raw_data" / task / "labelsTr"
            else:
                y_true_dir = workdir / labels_folder.replace(r"{fold}", str(fold))

            # evaluate
            metrics = evaluate_folder(
                y_det_dir=softmax_dir,
                y_true_dir=y_true_dir,
                subject_list=subject_list,
                pred_extensions=['_softmax.nii.gz'],
                y_det_postprocess_func=softmax_postprocessing_func,
                num_parallel_calls=5,
            )

            # save and show metrics
            metrics.save(metrics_path)
            print(f"Results for checkpoint {checkpoint}:")
            print(metrics)


def main():
    parser = argparse.ArgumentParser(description='Command Line Arguments')
    parser.add_argument("--task", type=str, default="Task2201_picai_baseline",
                        help="Task name of the nnU-Net experiment. Default: Task2201_picai_baseline")
    parser.add_argument("--trainer", type=str, default="nnUNetTrainerV2_Loss_FL_and_CE_checkpoints",
                        help="Trainer of the nnU-Net experiment. Default: nnUNetTrainerV2_Loss_FL_and_CE_checkpoints")
    parser.add_argument("--workdir", type=str, default="/workdir",
                        help="Path to the workdir where 'results' and 'nnUNet_raw_data' are stored. Default: /workdir")
    parser.add_argument("--task_dir", type=str, default="auto",
                        help="Path to the task directory (relative to the workdir). Optional, will be constucted " +
                             "for default nnU-Net folder structure")
    parser.add_argument("--checkpoints", type=str, nargs="+", default=["model_best"],
                        help="Which checkpoints to evaluate. Multiple checkpoints can be passed at once. Default: " +
                             "model_best")
    parser.add_argument("--folds", type=int, nargs="+", default=list(range(5)),
                        help="Which folds to evaluate. Multiple folds can be evaluated at once. Default: " +
                             "0, 1, 2, 3, 4  (all)")
    parser.add_argument("--predictions_folder", type=str, default=r"picai_pubtrain_predictions_{checkpoint}",
                        help="Folder with nnU-Net softmax predictions.")
    parser.add_argument("--labels_folder", type=str, default="auto",
                        help="Folder with annotations. Optional, will be constucted to labelsTr of the specified " +
                             "nnU-Net task.")
    parser.add_argument("--softmax_postprocessing_func", type=str, default="extract_lesion_candidates",
                        help="Function to post-process the softmax volumes. Default: extract lesion candidates " +
                             "using the Report-Guided Annotation repository. Use extract_lesion_candidates_cropped " +
                             "to set all predictions outside the central 20x384x384 voxels to zero.")
    parser.add_argument("--threshold", type=str, default="dynamic",
                        help="Threshold for lesion extraction from softmax predictions. " +
                             "Use dynamic-fast for quicker evaluation at almost equal performance." +
                             "Default: dynamic.")
    parser.add_argument("--metrics_fn", type=str, default=r"metrics-{checkpoint}-{fold}-{threshold}.json",
                        help=r"Filename to save metrics to. May contain {checkpoint} and {threshold} which are " +
                             r"auto-filled. Default: metrics-{checkpoint}-{threshold}.json")
    parser.add_argument("--splits", type=str, default="picai_pub_nnunet",
                        help="Splits for cross-validation. Available predefined splits: picai_pub, picai_pub_nnunet, " +
                             "picai_pubpriv, picai_pubpriv_nnunet. Alternatively, provide a path to a json file " +
                             "containing a dictiory with key 'subject_list', where `{fold}` in the path is replaced by " +
                             "the fold. Example: `/workdir/splits/val/fold_{fold}.json`. Default: picai_pub_nnunet.")
    args = parser.parse_args()

    # evaluate
    evaluate(
        task=args.task,
        trainer=args.trainer,
        workdir=args.workdir,
        task_dir=args.task_dir,
        checkpoints=args.checkpoints,
        folds=args.folds,
        softmax_postprocessing_func=args.softmax_postprocessing_func,
        threshold=args.threshold,
        metrics_fn=args.metrics_fn,
        splits=args.splits,
        predictions_folder=args.predictions_folder,
        labels_folder=args.labels_folder,
    )


if __name__ == "__main__":
    main()


#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from pathlib import Path


# read dataset configurations
train_splits = {}
for fold in range(5):
    with open(Path(__file__).parent / f"ds-config-train-fold-{fold}.json") as fp:
        ds_config = json.load(fp)
        train_splits[fold] = ds_config

valid_splits = {}
for fold in range(5):
    with open(Path(__file__).parent / f"ds-config-valid-fold-{fold}.json") as fp:
        ds_config = json.load(fp)
        valid_splits[fold] = ds_config

with open(Path(__file__).parent / "splits.json") as fp:
    nnunet_splits = json.load(fp)

# expose dataset configurations
__all__ = [
    "train_splits",
    "valid_splits",
    "nnunet_splits",
]



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import os

from picai_baseline.prepare_data import prepare_data
from picai_baseline.prepare_data_semi_supervised import \
    prepare_data_semi_supervised

if __name__ == '__main__':
    # Set up command line arguments
    argparser = argparse.ArgumentParser()
    subparsers = argparser.add_subparsers()

    # parse command line arguments
    parser = subparsers.add_parser('prepare_data_semi_supervised')
    parser.add_argument("--workdir", type=str, default=os.environ.get("workdir", "/workdir"),
                        help="Path to the working directory (default: /workdir, or the environment variable 'workdir')")
    parser.add_argument("--inputdir", type=str, default=os.environ.get("inputdir", "/input"),
                        help="Path to the input dataset (default: /input, or the environment variable 'inputdir')")
    parser.add_argument("--imagesdir", type=str, default="images",
                        help="Path to the images, relative to --inputdir (default: /input/images)")
    parser.add_argument("--labelsdir", type=str, default="picai_labels",
                        help="Path to the labels, relative to --inputdir (root of picai_labels) (default: /input/picai_labels)")
    parser.add_argument("--spacing", type=float, nargs="+", required=False,
                        help="Spacing to preprocess images to. Default: keep as-is.")
    parser.add_argument("--matrix_size", type=int, nargs="+", required=False,
                        help="Matrix size to preprocess images to. Default: keep as-is.")
    parser.add_argument("--preprocessing_kwargs", type=str, required=False,
                        help='Preprocessing kwargs to pass to the MHA2nnUNetConverter. " + \
                            "E.g.: `{"crop_only": true}`. Must be valid json.')
    parser.add_argument("--splits", type=str, default="picai_pub",
                        help="Splits to save for cross-validation. Available: picai_pub, picai_pubpriv.")
    parser.add_argument("--task", type=str, default="Task2203_picai_baseline",
                        help="Task name (default: Task2203_picai_baseline)")
    parser.set_defaults(func=prepare_data_semi_supervised)

    parser = subparsers.add_parser('prepare_data')
    parser.add_argument("--workdir", type=str, default=os.environ.get("workdir", "/workdir"),
                        help="Path to the working directory (default: /workdir, or the environment variable 'workdir')")
    parser.add_argument("--inputdir", type=str, default=os.environ.get("inputdir", "/input"),
                        help="Path to the input dataset (default: /input, or the environment variable 'inputdir')")
    parser.add_argument("--imagesdir", type=str, default="images",
                        help="Path to the images, relative to --inputdir (default: /input/images)")
    parser.add_argument("--labelsdir", type=str, default="picai_labels",
                        help="Path to the labels, relative to --inputdir (root of picai_labels) (default: /input/picai_labels)")
    parser.add_argument("--spacing", type=float, nargs="+", required=False,
                        help="Spacing to preprocess images to. Default: keep as-is.")
    parser.add_argument("--matrix_size", type=int, nargs="+", required=False,
                        help="Matrix size to preprocess images to. Default: keep as-is.")
    parser.add_argument("--preprocessing_kwargs", type=str, required=False,
                        help='Preprocessing kwargs to pass to the MHA2nnUNetConverter. " + \
                            "E.g.: `{"crop_only": true}`. Must be valid json.')
    parser.add_argument("--splits", type=str, default="picai_pub_nnunet",
                        help="Splits to save for cross-validation. Available: picai_pub_nnunet, picai_pubpriv_nnunet.")
    parser.add_argument("--task", type=str, default="Task2201_picai_baseline",
                        help="Task name (default: Task2201_picai_baseline)")
    parser.set_defaults(func=prepare_data)

    try:
        args = argparser.parse_args()
    except Exception as e:
        print(f"Parsing all arguments failed: {e}")
        print("Retrying with only the known arguments...")
        args, _ = argparser.parse_known_args()

    args.func(
        workdir=args.workdir,
        inputdir=args.inputdir,
        imagesdir=args.imagesdir,
        labelsdir=args.labelsdir,
        spacing=args.spacing,
        matrix_size=args.matrix_size,
        preprocessing_kwargs=args.preprocessing_kwargs,
        splits=args.splits,
        task=args.task,
    )
    print("Finished.")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np

from typing import Any, Optional
try:
    import numpy.typing as npt
except ImportError:  # pragma: no cover
    pass


def z_score_norm(image: "npt.NDArray[Any]", percentile: Optional[float] = None) -> "npt.NDArray[Any]":
    """
    Z-score normalization (mean=0; stdev=1), where intensities
    below or above the given percentile are discarded.
    [Ref: DLTK]

    Parameters:
    - image: N-dimensional image to be normalized
    - percentile: cutoff percentile (between 0-50)

    Returns:
    - scaled image
    """
    image = image.astype(np.float32)

    if percentile is not None:
        # clip distribution of intensity values
        lower_bnd = np.percentile(image, 100-percentile)
        upper_bnd = np.percentile(image, percentile)
        image = np.clip(image, lower_bnd, upper_bnd)

    # perform z-score normalization
    mean = np.mean(image)
    std = np.std(image)
    if std > 0:
        return (image - mean) / std
    else:
        return image * 0.



#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from picai_baseline.unet.training_setup.augmentations.multi_threaded_augmenter import MultiThreadedAugmenter
from picai_baseline.unet.training_setup.augmentations.single_threaded_augmenter import SingleThreadedAugmenter
from batchgenerators.transforms.abstract_transforms import Compose
from batchgenerators.transforms.color_transforms import BrightnessMultiplicativeTransform, ContrastAugmentationTransform, BrightnessTransform
from batchgenerators.transforms.color_transforms import GammaTransform
from batchgenerators.transforms.noise_transforms import GaussianNoiseTransform, GaussianBlurTransform
from batchgenerators.transforms.resample_transforms import SimulateLowResolutionTransform
from batchgenerators.transforms.spatial_transforms import SpatialTransform, MirrorTransform
from batchgenerators.transforms.utility_transforms import NumpyToTensor
try:
    from batchgenerators.dataloading.nondet_multi_threaded_augmenter import NonDetMultiThreadedAugmenter
except ImportError:
    NonDetMultiThreadedAugmenter = None

import numpy as np


default_3D_augmentation_params = {

    "do_elastic": False,
    "elastic_deform_alpha": (0., 900.),
    "elastic_deform_sigma": (9., 13.),
    "p_eldef": 0.2,
    "do_scaling": True,
    "scale_range": (0.7, 1.4),
    "independent_scale_factor_for_each_axis": False,
    "p_scale": 0.2,
    "do_rotation": True,
    "rotation_x": (-30. / 360 * 2. * np.pi, 30. / 360 * 2. * np.pi),
    "rotation_y": (-30. / 360 * 2. * np.pi, 30. / 360 * 2. * np.pi),
    "rotation_z": (-30. / 360 * 2. * np.pi, 30. / 360 * 2. * np.pi),
    "rotation_p_per_axis": 1,
    "p_rot": 0.2,
    "random_crop": False,
    "random_crop_dist_to_border": None,
    "do_gamma": True,
    "gamma_retain_stats": True,
    "gamma_range": (0.7, 1.5),
    "p_gamma": 0.3,
    "do_mirror": True,
    "mirror_axes": (0, 1, 2),
    "border_mode_data": "constant",
    "do_additive_brightness": False,
    "additive_brightness_p_per_sample": 0.15,
    "additive_brightness_p_per_channel": 0.5,
    "additive_brightness_mu": 0.0,
    "additive_brightness_sigma": 0.1,
}


def apply_augmentations(dataloader, params=default_3D_augmentation_params, patch_size=None, num_threads=1,
                        border_val_seg=-1, seeds_train=None, seeds_val=None, order_seg=1, order_data=3, disable=False,
                        pin_memory=False, use_multithreading=True, use_nondetMultiThreadedAugmenter: bool = False):
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------
    # initialize list for train-time transforms
    tr_transforms = []
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------
    if not disable:
        # morphological spatial transforms
        tr_transforms.append(SpatialTransform(patch_size,
                                              patch_center_dist_from_border=None,
                                              do_elastic_deform=params.get("do_elastic"),
                                              alpha=params.get("elastic_deform_alpha"),
                                              sigma=params.get("elastic_deform_sigma"),
                                              do_rotation=params.get("do_rotation"),
                                              angle_x=params.get("rotation_x"),
                                              angle_y=params.get("rotation_y"),
                                              angle_z=params.get("rotation_z"),
                                              p_rot_per_axis=params.get("rotation_p_per_axis"),
                                              do_scale=params.get("do_scaling"),
                                              scale=params.get("scale_range"),
                                              border_mode_data=params.get("border_mode_data"),
                                              border_cval_data=0,
                                              order_data=order_data,
                                              border_mode_seg="constant",
                                              border_cval_seg=border_val_seg,
                                              order_seg=order_seg,
                                              random_crop=params.get("random_crop"),
                                              p_el_per_sample=params.get("p_eldef"),
                                              p_scale_per_sample=params.get("p_scale"),
                                              p_rot_per_sample=params.get("p_rot"),
                                              independent_scale_for_each_axis=params.get("independent_scale_factor_for_each_axis")))
        # -------------------------------------------------------------------------------------------------------------------------------------------------------------
        # intensity transforms
        tr_transforms.append(GaussianNoiseTransform(p_per_sample=0.1))
        tr_transforms.append(GaussianBlurTransform((0.5, 1.), different_sigma_per_channel=True, p_per_sample=0.2, p_per_channel=0.5))
        tr_transforms.append(BrightnessMultiplicativeTransform(multiplier_range=(0.75, 1.25), p_per_sample=0.15))

        if params.get("do_additive_brightness"):
            tr_transforms.append(BrightnessTransform(params.get("additive_brightness_mu"),
                                                     params.get("additive_brightness_sigma"), True,
                                                     p_per_sample=params.get("additive_brightness_p_per_sample"),
                                                     p_per_channel=params.get("additive_brightness_p_per_channel")))

        tr_transforms.append(ContrastAugmentationTransform(p_per_sample=0.15))
        tr_transforms.append(SimulateLowResolutionTransform(zoom_range=(0.5, 1), per_channel=True, p_per_channel=0.5,
                                                            order_downsample=0,  order_upsample=3, p_per_sample=0.25,
                                                            ignore_axes=None))
        tr_transforms.append(
            GammaTransform(params.get("gamma_range"), True, True, retain_stats=params.get("gamma_retain_stats"),
                           p_per_sample=0.1))  # inverted gamma

        if params.get("do_gamma"):
            tr_transforms.append(
                GammaTransform(params.get("gamma_range"), False, True, retain_stats=params.get("gamma_retain_stats"),
                               p_per_sample=params["p_gamma"]))
        # ------------------------------------------------------------------------------------------------------------------------------------------------------------
        # flipping transform (reserved for last in order)
        if params.get("do_mirror") or params.get("mirror"):
            tr_transforms.append(MirrorTransform(params.get("mirror_axes")))
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------
    # convert NumPy -> PyTorch tensors
    tr_transforms.append(NumpyToTensor(['data', 'seg'], 'float'))
    tr_transforms = Compose(tr_transforms)
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------
    # multi-threaded augmenter (non-deterministic, if available)
    if use_multithreading and num_threads > 1:
        if use_nondetMultiThreadedAugmenter:
            if NonDetMultiThreadedAugmenter is None:
                raise RuntimeError('NonDetMultiThreadedAugmenter is not yet available')
            batchgenerator_train = NonDetMultiThreadedAugmenter(dataloader, tr_transforms, num_threads,
                                                                num_threads, seeds=seeds_train,
                                                                pin_memory=pin_memory)
        else:
            batchgenerator_train = MultiThreadedAugmenter(dataloader, tr_transforms, num_threads,
                                                          num_threads, seeds=seeds_train, 
                                                          pin_memory=pin_memory)
    else:
        batchgenerator_train = SingleThreadedAugmenter(dataloader, tr_transforms)
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------
    return batchgenerator_train
    # ----------------------------------------------------------------------------------------------------------------------------------------------------------------



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import os
import torch


def compute_spec_for_run(args, train_device='cuda'):
    """Register GPU/CPU specifications"""
    device = torch.device(train_device)
    
    # os.sched_getaffinity(0) is not supported by all operating systems
    try:
        args.num_threads = np.min([len(os.sched_getaffinity(0)), args.max_threads])
    except:
        args.num_threads = args.max_threads
    
    print("-"*100)
    print("GPUs:", torch.cuda.get_device_name(0))
    print("CPU Threads:", args.num_threads)
    return device, args



"""
Minimal subset of DIAGNijmegen/msk-tiger/io.py, as required for the nnunet wrapper.
"""

import hashlib
import json
import subprocess
import sys
from collections import OrderedDict
from pathlib import Path
from typing import (
    Any,
    Union,
)

# We generally except strings and pathlib paths for file and folder names
PathLike = Union[str, Path]

# --------------------------------------------------------------------------------


class JSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, Path):
            return o.as_posix()
        else:
            return super().default(o)


def read_json(filename: PathLike, *, ordered_dict: bool = True, **kwargs):
    """Reads a json file"""
    if ordered_dict:
        kwargs["object_pairs_hook"] = OrderedDict

    with Path(filename).open() as fp:
        return json.load(fp, **kwargs)


def write_json(
    filename: PathLike,
    data: Any,
    *,
    encoding: str = "UTF-8",
    make_dirs: bool = True,
    **kwargs,
):
    """Dumps an object into a json file, using pretty printing and UTF-8 encoding by default"""
    jsonfile = Path(filename)

    # Ensure that the directory exists
    if make_dirs:
        jsonfile.parent.mkdir(parents=True, exist_ok=True)

    # Write data into json file
    args = {"sort_keys": False, "indent": 2, "cls": JSONEncoder}
    args.update(kwargs)

    with jsonfile.open("w", encoding=encoding) as fp:
        json.dump(data, fp, **args)


# --------------------------------------------------------------------------------


def refresh_file_list(path: PathLike):
    """Update the cached file list of directories on a network share

    On network shares (chansey in particular), files are sometimes reported missing even though they
    exist. This has to do with caching issues and can be fixed by running "ls" or similar commands
    in the parent directory of these files.
    """
    path = Path(path)  # make sure path is a OS-specific path object
    if sys.platform == "win32":
        cmd = ["cmd", "/c", "dir", str(path)]
    else:
        cmd = ["ls", str(path)]

    try:
        subprocess.check_call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        raise OSError(f'Could not refresh file list of directory "{path}"') from e


def path_exists(path: PathLike) -> bool:
    """Checks whether the file or directory exists

    Unlike checks via os.path or pathlib, this check works reliably also on network shares where the
    content of directories might be cached.
    """
    path = Path(path)

    # Refresh content of parent (folder in which the object of interest is stored)
    try:
        refresh_file_list(path.parent)
    except OSError:
        # If the parent directory does not exist, refreshing the file list will fail, but that's okay
        pass

    # Now we can check if the specific file/directory exists
    if path.exists():
        # If the object was a directory itself, we also ask for a fresh list of it's content
        if path.is_dir() and path != path.parent:
            refresh_file_list(path)
        return True
    else:
        return False


def checksum(file: PathLike, algorithm: str = "sha256", chunk_size: int = 4096) -> str:
    """Computes the checksum of a file using a hashing algorithm"""
    file = Path(file)
    if file.exists() and not file.is_file():
        raise ValueError(f"Checksum can be computed only for files, {file} is not a file")

    h = hashlib.new(algorithm)
    with file.open("rb") as fp:
        for chunk in iter(lambda: fp.read(chunk_size), b""):
            h.update(chunk)
    return h.hexdigest()



# Copyright 2021 Division of Medical Image Computing, German Cancer Research Center (DKFZ)
# and Applied Computer Vision Lab, Helmholtz Imaging Platform
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
from typing import List, Union
import threading
from multiprocessing import Process, Queue
from queue import Queue as thrQueue
import numpy as np
import sys
import logging
from multiprocessing import Event
from time import sleep, time
from threadpoolctl import threadpool_limits

try:
    import torch
except ImportError:
    torch = None


def producer(queue, data_loader, transform, thread_id, seed, abort_event, wait_time: float = 0.02):
    np.random.seed(seed)
    data_loader.set_thread_id(thread_id)
    item = None

    try:
        while True:
            # check if abort event was set
            if not abort_event.is_set():
                # print("worker %d event not set" % thread_id)
                if item is None:
                    try:
                        item = next(data_loader)
                        if transform is not None:
                            item = transform(**item)
                    except StopIteration:
                        item = "end"

                if not queue.full():
                    queue.put(item)
                    item = None
                else:
                    sleep(wait_time)
            else:
                # print("worder %d event is now set, exiting" % thread_id)
                return
    except KeyboardInterrupt:
        abort_event.set()
        return
    except Exception as e:
        print("Exception in background worker %d:\n" % thread_id, e)
        traceback.print_exc()
        abort_event.set()
        return


def results_loop(in_queues: List[Queue], out_queue: thrQueue, abort_event: Event, pin_memory: bool,
                 gpu: Union[int, None], wait_time: float, worker_list: list):
    do_pin_memory = torch is not None and pin_memory and gpu is not None and torch.cuda.is_available()

    if do_pin_memory:
        print('using pin_memory on device', gpu)
        torch.cuda.set_device(gpu)

    item = None
    queue_ctr = 0
    end_ctr = 0

    while True:
        # if abort_event is set we need to clean up. This is where it hangs sometimes so it makes sense to drain all
        # the incoming queues and ignore all the errors occuring during this process.
        try:
            if abort_event.is_set():
                return

            # check if all workers are still alive
            if not all([i.is_alive() for i in worker_list]):
                abort_event.set()

            # if we don't have an item we need to fetch it first. If the queue we want to get it from it empty, try
            # again later
            if item is None:
                current_queue = in_queues[queue_ctr % len(in_queues)]
                if not current_queue.empty():
                    # get the item
                    item = current_queue.get()
                    # if we do pin memory, do it now, otherwise skip this
                    if do_pin_memory:
                        if isinstance(item, dict):
                            for k in item.keys():
                                if isinstance(item[k], torch.Tensor):
                                    item[k] = item[k].pin_memory()
                    queue_ctr += 1

                    if isinstance(item, str) and item == 'end':
                        end_ctr += 1
                    if end_ctr == len(in_queues):
                        end_ctr = 0
                        queue_ctr = 0

                else:
                    sleep(wait_time)
                    continue

            # we only arrive here if item is not None. Now put item in to the out_queue
            if not out_queue.full():
                out_queue.put(item)
                item = None
            else:
                sleep(wait_time)
                continue
        except KeyboardInterrupt:
            abort_event.set()
            raise KeyboardInterrupt


class MultiThreadedAugmenter(object):
    """ Makes your pipeline multi threaded. Yeah!
    If seeded we guarantee that batches are retunred in the same order and with the same augmentation every time this
    is run. This is realized internally by using une queue per worker and querying the queues one ofter the other.
    Args:
        data_loader (generator or DataLoaderBase instance): Your data loader. Must have a .next() function and return
        a dict that complies with our data structure
        transform (Transform instance): Any of our transformations. If you want to use multiple transformations then
        use our Compose transform! Can be None (in that case no transform will be applied)
        num_processes (int): number of processes
        num_cached_per_queue (int): number of batches cached per process (each process has its own
        multiprocessing.Queue). We found 2 to be ideal.
        seeds (list of int): one seed for each worker. Must have len(num_processes).
        If None then seeds = range(num_processes)
        pin_memory (bool): set to True if all torch tensors in data_dict are to be pinned. Pytorch only.
        timeout (int): How long do we wait for the background workers to do stuff? If timeout seconds have passed and
        self.__get_next_item still has not gotten an item from the workers we will perform a check whether all
        background workers are still alive. If all are alive we wait, if not we set the abort flag.
        wait_time (float): set this to be lower than the time you need per iteration. Don't set this to 0,
        that will come with a performance penalty. Default is 0.02 which will be fine for 50 iterations/s
    """

    def __init__(self, data_loader, transform, num_processes, num_cached_per_queue=2, seeds=None, pin_memory=False,
                 timeout=10, wait_time=0.02):
        self.timeout = timeout
        self.pin_memory = pin_memory
        self.transform = transform
        if seeds is not None:
            assert len(seeds) == num_processes
        else:
            seeds = [None] * num_processes
        self.seeds = seeds
        self.generator = data_loader
        self.num_processes = num_processes
        self.num_cached_per_queue = num_cached_per_queue
        self._queues = []
        self._processes = []
        self._end_ctr = 0
        self._queue_ctr = 0
        self.pin_memory_thread = None
        self.pin_memory_queue = None
        self.abort_event = Event()
        self.wait_time = wait_time
        self.was_initialized = False

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __get_next_item(self):
        item = None

        while item is None:
            if self.abort_event.is_set():
                self._finish()

            if not self.pin_memory_queue.empty():
                item = self.pin_memory_queue.get()
            else:
                sleep(self.wait_time)

        return item

    def __next__(self):
        if not self.was_initialized:
            self._start()

        try:
            item = self.__get_next_item()

            while isinstance(item, str) and (item == "end"):
                self._end_ctr += 1
                if self._end_ctr == self.num_processes:
                    self._end_ctr = 0
                    self._queue_ctr = 0
                    logging.debug("MultiThreadedGenerator: finished data generation")
                    raise StopIteration

                item = self.__get_next_item()

            return item

        except KeyboardInterrupt:
            logging.error("MultiThreadedGenerator: caught exception: {}".format(sys.exc_info()))
            self.abort_event.set()
            self._finish()
            raise KeyboardInterrupt

    def _start(self):
        if not self.was_initialized:
            self._finish()
            self.abort_event.clear()

            logging.debug("starting workers")
            self._queue_ctr = 0
            self._end_ctr = 0

            if hasattr(self.generator, 'was_initialized'):
                self.generator.was_initialized = False

            with threadpool_limits(limits=1, user_api="blas"):
                for i in range(self.num_processes):
                    self._queues.append(Queue(self.num_cached_per_queue))
                    self._processes.append(Process(target=producer, args=(
                        self._queues[i], self.generator, self.transform, i, self.seeds[i], self.abort_event)))
                    self._processes[-1].daemon = True
                    self._processes[-1].start()

            if torch is not None and torch.cuda.is_available():
                gpu = torch.cuda.current_device()
            else:
                gpu = None

            # more caching = more performance. But don't cache too much or your RAM will hate you
            self.pin_memory_queue = thrQueue(max(3, self.num_cached_per_queue * self.num_processes // 2))

            self.pin_memory_thread = threading.Thread(target=results_loop, args=(
                self._queues, self.pin_memory_queue, self.abort_event, self.pin_memory, gpu, self.wait_time,
                self._processes))

            self.pin_memory_thread.daemon = True
            self.pin_memory_thread.start()

            self.was_initialized = True
        else:
            logging.debug("MultiThreadedGenerator Warning: start() has been called but it has already been "
                          "initialized previously")

    def _finish(self, timeout=10):
        self.abort_event.set()

        start = time()
        while self.pin_memory_thread is not None and self.pin_memory_thread.is_alive() and start + timeout > time():
            sleep(0.2)

        if len(self._processes) != 0:
            logging.debug("MultiThreadedGenerator: shutting down workers...")
            [i.terminate() for i in self._processes]

            for i, p in enumerate(self._processes):
                self._queues[i].close()
                self._queues[i].join_thread()

            self._queues = []
            self._processes = []
            self._queue = None
            self._end_ctr = 0
            self._queue_ctr = 0

            del self.pin_memory_queue
        self.was_initialized = False

    def restart(self):
        self._finish()
        self._start()

# << Not Required in Dockerized Setup >>
    def __del__(self):
        logging.debug("MultiThreadedGenerator: destructor was called")
        self._finish()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


unet_hyperparam = {
    'batch_size': 8,
    'model_strides': [(2, 2, 2), (1, 2, 2), (1, 2, 2), (1, 2, 2), (2, 2, 2)],
    'model_features': [32, 64, 128, 256, 512, 1024]
}


def get_default_hyperparams(args):
    """Retrieve default hyperparameters for given neural network architecture"""

    # used for inference
    if isinstance(args, dict):
        if args['model_type'] == 'unet':
            args['model_strides'] = unet_hyperparam['model_strides']
            args['model_features'] = unet_hyperparam['model_features']
        args = type('_', (object,), args)

    # used at train-time
    else:
        if args.model_type == 'unet':
            args.batch_size = unet_hyperparam['batch_size']
            args.model_strides = unet_hyperparam['model_strides']
            args.model_features = unet_hyperparam['model_features']
    return args



# Copyright 2021 Division of Medical Image Computing, German Cancer Research Center (DKFZ)
# and Applied Computer Vision Lab, Helmholtz Imaging Platform
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class SingleThreadedAugmenter(object):
    """
    Use this for debugging custom transforms. It does not use a background thread and you can therefore easily debug
    into your augmentations. This should not be used for training. If you want a generator that uses (a) background
    process(es), use MultiThreadedAugmenter.
    Args:
        data_loader (generator or DataLoaderBase instance): Your data loader. Must have a .next() function and return
        a dict that complies with our data structure
        transform (Transform instance): Any of our transformations. If you want to use multiple transformations then
        use our Compose transform! Can be None (in that case no transform will be applied)
    """

    def __init__(self, data_loader, transform):
        self.data_loader = data_loader
        self.transform = transform

    def __iter__(self):
        return self

    def __next__(self):
        item = next(self.data_loader)
        if self.transform is not None:
            item = self.transform(**item)
        return item

    def next(self):
        return self.__next__()



from picai_baseline.splits import export_splits
from picai_baseline.splits.picai_debug import (nnunet_splits, train_splits,
                                               valid_splits)

if __name__ == "__main__":
    export_splits(
        train_splits=train_splits,
        valid_splits=valid_splits,
        nnunet_splits=nnunet_splits,
    )



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torch
import torch.nn as nn
import torch.nn.functional as F


class FocalLoss(nn.Module):
    """Focal loss function for binary segmentation."""

    def __init__(self, alpha=1, gamma=2, num_classes=2, reduction="sum"):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.num_classes = num_classes
        self.reduction = reduction

    def forward(self, inputs, targets):
        inputs = torch.sigmoid(inputs)
        targets = F.one_hot(targets, num_classes=self.num_classes).float()
        targets = torch.moveaxis(targets, (0, 1, 2, 3, 4), (0, 2, 3, 4, 1))
        ce_loss = F.binary_cross_entropy(inputs, targets, reduction="none")
        p_t = (inputs[-1] * targets[-1]) + ((1 - inputs[-1]) * (1 - targets[-1]))
        loss = ce_loss * ((1 - p_t) ** self.gamma)

        if self.alpha >= 0:
            alpha_t = self.alpha * targets[-1] + (1 - self.alpha) * (1 - targets[-1])
            loss = alpha_t * loss

        if self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "sum":
            loss = loss.sum()

        return loss



import torch
from batchgenerators.augmentations.spatial_transformations import augment_zoom, augment_resize


# Horizontal Flip for 5D Tensors (B,C,W,H,D)
def hflip_tensor(tensor):
    return tensor.flip(-3)

# Simulate Low-Res Scan for 5D Tensors (B,C,H,W,D)


def sim_low_res_tensor(tensor):
    img_npy = tensor.detach().numpy()[0]
    og_shape = img_npy.shape

    # Downsample Image by 12.5%
    img_npy, _ = augment_zoom(img_npy, None, (0.875, 0.875, 1.000))

    # Upsample to Original Size via Nearest Neighbor Interpolation
    img_npy, _ = augment_resize(img_npy, None, og_shape[0], order=0)

    return torch.unsqueeze(torch.from_numpy(img_npy.copy()), 0)



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import ast

import torch
from picai_baseline.unet.training_setup.augmentations.nnUNet_DA import \
    apply_augmentations
from picai_baseline.unet.training_setup.callbacks import (
    optimize_model, resume_or_restart_training, validate_model)
from picai_baseline.unet.training_setup.compute_spec import \
    compute_spec_for_run
from picai_baseline.unet.training_setup.data_generator import prepare_datagens
from picai_baseline.unet.training_setup.default_hyperparam import \
    get_default_hyperparams
from picai_baseline.unet.training_setup.loss_functions.focal import FocalLoss
from picai_baseline.unet.training_setup.neural_network_selector import \
    neural_network_for_run
from torch.utils.tensorboard import SummaryWriter


def main():
    # command line arguments for hyperparameters and I/O paths
    parser = argparse.ArgumentParser(description='Command Line Arguments for Training Script')

    # data I/0 + experimental setup
    parser.add_argument('--max_threads', type=int, default=12,
                        help="Max threads/workers for data loaders")
    parser.add_argument('--validate_n_epochs', type=int, default=10,               
                        help="Trigger validation every N epochs")
    parser.add_argument('--validate_min_epoch', type=int, default=50,               
                        help="Trigger validation after minimum N epochs")
    parser.add_argument('--export_best_model', type=int, default=1,                
                        help="Export model checkpoints")
    parser.add_argument('--resume_training', type=str, default=1,                
                        help="Resume training model, if checkpoint exists")
    parser.add_argument('--weights_dir', type=str, required=True,            
                        help="Path to export model checkpoints")
    parser.add_argument('--overviews_dir', type=str, required=True,            
                        help="Base path to training/validation data sheets")
    parser.add_argument('--folds', type=int, nargs='+', required=True, 
                        help="Folds selected for training/validation run")

    # training hyperparameters
    parser.add_argument('--image_shape', type=int, nargs="+", default=[20, 256, 256],   
                        help="Input image shape (z, y, x)")
    parser.add_argument('--num_channels', type=int, default=3,                
                        help="Number of input channels/sequences")
    parser.add_argument('--num_classes', type=int, default=2,                
                        help="Number of classes at train-time")
    parser.add_argument('--num_epochs', type=int, default=100,              
                        help="Number of training epochs")
    parser.add_argument('--base_lr', type=float, default=0.001,            
                        help="Learning rate")
    parser.add_argument('--focal_loss_gamma', type=float, default=1.0,              
                        help="Focal Loss gamma value")
    parser.add_argument('--enable_da', type=int, default=1,                
                        help="Enable data augmentation")

    # neural network-specific hyperparameters
    parser.add_argument('--model_type', type=str, default='unet',                                                    
                        help="Neural network: architectures")
    parser.add_argument('--model_strides', type=str, default='[(2, 2, 2), (1, 2, 2), (1, 2, 2), (1, 2, 2), (2, 2, 2)]', 
                        help="Neural network: convolutional strides (as string representation)")
    parser.add_argument('--model_features', type=str, default='[32, 64, 128, 256, 512, 1024]',                           
                        help="Neural network: number of encoder channels (as string representation)")
    parser.add_argument('--batch_size', type=int, default=8,                                                         
                        help="Mini-batch size")
    parser.add_argument('--use_def_model_hp', type=int, default=1,                                                         
                        help="Use default set of model-specific hyperparameters")

    args = parser.parse_args()
    args.model_strides = ast.literal_eval(args.model_strides)
    args.model_features = ast.literal_eval(args.model_features)

    # retrieve default set of hyperparam (architecture, batch size) for given neural network
    if bool(args.use_def_model_hp):
        args = get_default_hyperparams(args)

    # for each fold
    for f in args.folds:
        # --------------------------------------------------------------------------------------------------------------------------
        # GPU/CPU specifications
        device, args = compute_spec_for_run(args=args)

        # derive dataLoaders
        train_gen, valid_gen, class_weights = prepare_datagens(args=args, fold_id=f)

        # integrate data augmentation pipeline from nnU-Net
        train_gen = apply_augmentations(
            dataloader=train_gen,
            num_threads=args.num_threads,
            disable=(not bool(args.enable_da))
        )
        
        # initialize multi-threaded augmenter in background
        train_gen.restart()

        # model definition
        model = neural_network_for_run(args=args, device=device)

        # loss function + optimizer
        loss_func = FocalLoss(alpha=class_weights[-1], gamma=args.focal_loss_gamma).to(device)
        optimizer = torch.optim.Adam(params=model.parameters(), lr=args.base_lr, amsgrad=True)
        # --------------------------------------------------------------------------------------------------------------------------
        # training loop
        writer = SummaryWriter()

        # resume or restart training model, based on whether checkpoint exists
        model, optimizer, tracking_metrics = resume_or_restart_training(
            model=model, optimizer=optimizer,
            device=device, args=args, fold_id=f
        )

        # for each epoch
        for epoch in range(tracking_metrics['start_epoch'], args.num_epochs):

            # optimize model x N training steps + update learning rate
            model.train()
            tracking_metrics['epoch'] = epoch

            model, optimizer, train_gen, tracking_metrics, writer = optimize_model(
                model=model, optimizer=optimizer, loss_func=loss_func, train_gen=train_gen,
                args=args, tracking_metrics=tracking_metrics, device=device, writer=writer
            )

            # ----------------------------------------------------------------------------------------------------------------------
            # for each round of validation
            if ((epoch+1) % args.validate_n_epochs == 0) and ((epoch+1) >= args.validate_min_epoch):

                # validate model per N epochs + export model weights
                model.eval()
                with torch.no_grad():  # no gradient updates during validation

                    model, optimizer, valid_gen, tracking_metrics, writer = validate_model(
                        model=model, optimizer=optimizer, valid_gen=valid_gen, args=args,
                        tracking_metrics=tracking_metrics, device=device, writer=writer
                    )

        # --------------------------------------------------------------------------------------------------------------------------
        print(
            f"Training Complete! Peak Validation Ranking Score: {tracking_metrics['best_metric']:.4f} "
            f"@ Epoch: {tracking_metrics['best_metric_epoch']}")
        writer.close()
        # --------------------------------------------------------------------------------------------------------------------------


if __name__ == '__main__':
    main()



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import SimpleITK as sitk

from picai_baseline.splits.picai import nnunet_splits as picai_pub_splits
from picai_baseline.splits.picai_nnunet import \
    nnunet_splits as picai_pub_nnunet_splits
from picai_baseline.splits.picai_pubpriv import \
    nnunet_splits as picai_pubpriv_splits
from picai_baseline.splits.picai_pubpriv_nnunet import \
    nnunet_splits as picai_pubpriv_nnunet_splits


def main(
    preprocessed_data_path: Union[Path, str] = Path('/workdir/nnUNet_raw_data/Task2201_picai_baseline/'),
    overviews_path: Union[Path, str] = Path('/workdir/results/UNet/overviews/'),
    splits: Optional[Dict[str, List[str]]] = None,
):
    """Create overviews of the training data."""

    if isinstance(splits, str):
        # select splits
        splits = {
            "picai_pub": picai_pub_splits,
            "picai_pubpriv": picai_pubpriv_splits,
            "picai_pub_nnunet": picai_pub_nnunet_splits,
            "picai_pubpriv_nnunet": picai_pubpriv_nnunet_splits,
        }[args.splits]

    if splits is None:
        raise ValueError("No splits provided!")

    # create directory to store overviews
    overviews_path.mkdir(parents=True, exist_ok=True)

    # iterate over each cross-validation fold
    for fold, nnunet_fold in enumerate(splits):

        # iterate over train and validation splits
        for split, nnunet_split in nnunet_fold.items():
            print(f"Preparing fold {fold} ({split})..")

            # initialize list of fields to collect for each split of each fold
            overview = {
                'pat_ids': [],
                'study_ids': [],
                'image_paths': [],
                'label_paths': [],
                'case_label': [],
                'ratio_csPCa_bg': []
            }

            # iterate over each training/validation case
            for subject_id in nnunet_split:
                patient_id, study_id = subject_id.split('_')

                # skip cases where preprocessing failed
                if not (preprocessed_data_path / 'labelsTr' / f'{subject_id}.nii.gz').exists():
                    print(f"Skipping {subject_id}, case not found (likely because preprocessing failed)!")
                    continue

                # load annotation
                lbl = sitk.GetArrayFromImage(sitk.ReadImage(str(preprocessed_data_path / 'labelsTr' / f'{subject_id}.nii.gz')))

                overview['pat_ids'] += [patient_id]
                overview['study_ids'] += [study_id]
                overview['image_paths'] += [[
                    str((preprocessed_data_path / 'imagesTr' / f'{subject_id}_0000.nii.gz').as_posix()),
                    str((preprocessed_data_path / 'imagesTr' / f'{subject_id}_0001.nii.gz').as_posix()),
                    str((preprocessed_data_path / 'imagesTr' / f'{subject_id}_0002.nii.gz').as_posix()),
                ]]
                overview['label_paths'] += [str((preprocessed_data_path / 'labelsTr' / f'{subject_id}.nii.gz').as_posix())]
                overview['case_label'] += [float(np.max(lbl))]
                overview['ratio_csPCa_bg'] += [float(np.sum(lbl)/np.size(lbl))]

            # save overview
            with open(overviews_path / f'PI-CAI_{split}-fold-{fold}.json', 'w') as fp:
                json.dump(overview, fp, indent=4)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Command Line Arguments')
    parser.add_argument("--task", type=str, default="Task2201_picai_baseline",
                        help="Task name of the experiment. Default: Task2201_picai_baseline")
    parser.add_argument("--workdir", type=str, default="/workdir",
                        help="Path to the workdir where 'results' and 'nnUNet_raw_data' are stored. Default: /workdir")
    parser.add_argument("--preprocessed_data_path", type=str, default="nnUNet_raw_data/{task}",
                        help="Path to the preprocessed data, relative to the workdir. Default: /workdir/nnUNet_raw_data/{task}")
    parser.add_argument("--overviews_path", type=str, default="results/UNet/overviews/{task}",
                        help="Path to the overviews, relative to the workdir. Default: /workdir/results/UNet/overviews/{task}")
    parser.add_argument("--splits", type=str, default="picai_pub_nnunet",
                        help="Splits for cross-validation. Available: picai_pub, picai_pub_nnunet, picai_pubpriv, " +
                             "picai_pubpriv_nnunet.")
    args = parser.parse_args()

    # paths
    workdir = Path(args.workdir)
    preprocessed_data_path = workdir / args.preprocessed_data_path.replace("{task}", args.task)
    overviews_path = workdir / args.overviews_path.replace("{task}", args.task)

    # evaluate
    main(
        preprocessed_data_path=preprocessed_data_path,
        overviews_path=overviews_path,
        splits=args.splits,
    )
    print("Finished.")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from copy import deepcopy

from picai_baseline.splits import subject_list_annotated_pubpriv
from picai_baseline.splits.picai_pubpriv import (nnunet_splits, train_splits,
                                                 valid_splits)

# read dataset configurations
train_splits = deepcopy(train_splits)  # copy to avoid modifying the original
for fold, split in train_splits.items():
    split['subject_list'] = sorted(list(set(split['subject_list']) & set(subject_list_annotated_pubpriv)))

valid_splits = deepcopy(valid_splits)  # copy to avoid modifying the original
for fold, split in valid_splits.items():
    split['subject_list'] = sorted(list(set(split['subject_list']) & set(subject_list_annotated_pubpriv)))

nnunet_splits = deepcopy(nnunet_splits)  # copy to avoid modifying the original
for split in nnunet_splits:
    split['train'] = sorted(list(set(split['train']) & set(subject_list_annotated_pubpriv)))
    split['val'] = sorted(list(set(split['val']) & set(subject_list_annotated_pubpriv)))

# expose dataset configurations
__all__ = [
    "train_splits",
    "valid_splits",
    "nnunet_splits",
]



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import json
import os
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Union

import SimpleITK as sitk
from picai_prep import MHA2nnUNetConverter
from picai_prep.data_utils import atomic_image_write
from picai_prep.examples.mha2nnunet.picai_archive import \
    generate_mha2nnunet_settings
from tqdm import tqdm

from picai_baseline.splits.picai import nnunet_splits as picai_splits
from picai_baseline.splits.picai_pubpriv import \
    nnunet_splits as picai_pubpriv_splits

"""
Script to prepare PI-CAI data into the nnUNet raw data format
For documentation, please see:
https://github.com/DIAGNijmegen/picai_baseline#data-preparation
"""

def preprocess_picai_annotation(lbl: sitk.Image) -> sitk.Image:
    """Binarize the granular ISUP ≥ 2 annotations"""
    lbl_arr = sitk.GetArrayFromImage(lbl)

    # convert granular PI-CAI csPCa annotation to binary csPCa annotation
    lbl_arr = (lbl_arr >= 1).astype('uint8')

    # convert label back to SimpleITK
    lbl_new: sitk.Image = sitk.GetImageFromArray(lbl_arr)
    lbl_new.CopyInformation(lbl)
    return lbl_new


def prepare_data_semi_supervised(
    workdir: Union[Path, str] = "/workdir",
    inputdir: Union[Path, str] = "/input",
    imagesdir: str = "images",
    labelsdir: str = "picai_labels",
    spacing: Optional[Iterable[float]] = None,
    matrix_size: Optional[Iterable[int]] = None,
    preprocessing_kwargs: Optional[Dict[str, Any]] = None,
    splits: str = "picai_pub",
    task: str = "Task2203_picai_baseline",
):
    """
    Script to prepare PI-CAI data into the nnUNet raw data format
    For documentation, please see:
    https://github.com/DIAGNijmegen/picai_baseline#data-preparation
    """

    # prepare preprocessing kwargs
    if preprocessing_kwargs is None or preprocessing_kwargs == "":
        preprocessing_kwargs = {}
    elif isinstance(preprocessing_kwargs, str):
        preprocessing_kwargs = json.loads(preprocessing_kwargs)
    if not isinstance(preprocessing_kwargs, dict):
        raise ValueError("preprocessing_kwargs must be a dict or None")
    if spacing:
        if "spacing" in preprocessing_kwargs:
            raise ValueError("Cannot specify both --spacing and --preprocessing_kwargs['spacing']")
        preprocessing_kwargs["spacing"] = spacing
    if matrix_size:
        if "matrix_size" in preprocessing_kwargs:
            raise ValueError("Cannot specify both --matrix_size and --preprocessing_kwargs['matrix_size']")
        preprocessing_kwargs["matrix_size"] = matrix_size

    # select splits
    splits = {
        "picai_pub": picai_splits,
        "picai_pubpriv": picai_pubpriv_splits,
    }[splits]

    # parse paths
    workdir = Path(workdir)
    inputdir = Path(inputdir)
    imagesdir = Path(inputdir / imagesdir)
    labelsdir = Path(inputdir / labelsdir)

    # relative paths
    annotations_dir_human_expert = labelsdir / "csPCa_lesion_delineations/human_expert/resampled/"
    annotations_dir_ai_derived = labelsdir / "csPCa_lesion_delineations/AI/Bosma22a/"
    annotations_dir = labelsdir / "csPCa_lesion_delineations/combined/"
    mha2nnunet_settings_path = workdir / "mha2nnunet_settings" / (task + ".json")
    nnUNet_raw_data_path = workdir / "nnUNet_raw_data"
    nnUNet_task_dir = nnUNet_raw_data_path / task
    nnUNet_dataset_json_path = nnUNet_task_dir / "dataset.json"
    nnUNet_splits_path = nnUNet_task_dir / "splits.json"

    # prepare annotations folder with human-expert and AI-derived annotations
    annotations_dir.mkdir(parents=True, exist_ok=True)
    if len(os.listdir(annotations_dir)) == 1500:
        print("Annotations folder already prepared, skipping..")
    else:
        print("Preparing annotations folder with human-expert and AI-derived annotations")
        for fn in tqdm(os.listdir(annotations_dir_human_expert), desc="Copying human expert annotations"):
            path_src = annotations_dir_human_expert / fn
            path_dst = annotations_dir / fn
            if fn.endswith(".nii.gz") and not path_dst.exists():
                lbl = sitk.ReadImage(str(path_src))
                atomic_image_write(lbl, path_dst)
        for fn in tqdm(os.listdir(annotations_dir_ai_derived), desc="Copying AI-derived annotations"):
            path_src = annotations_dir_ai_derived / fn
            path_dst = annotations_dir / fn
            if fn.endswith(".nii.gz") and not path_dst.exists():
                lbl = sitk.ReadImage(str(path_src))
                atomic_image_write(lbl, path_dst)

    if mha2nnunet_settings_path.exists():
        print(f"Found mha2nnunet settings at {mha2nnunet_settings_path}, skipping..")
    else:
        # generate mha2nnunet conversion plan
        Path(mha2nnunet_settings_path.parent).mkdir(parents=True, exist_ok=True)
        generate_mha2nnunet_settings(
            archive_dir=imagesdir,
            annotations_dir=annotations_dir,
            output_path=mha2nnunet_settings_path,
            task=task,
        )

        # read mha2nnunet_settings
        with open(mha2nnunet_settings_path) as fp:
            mha2nnunet_settings = json.load(fp)

        # note: modify preprocessing settings here
        mha2nnunet_settings["preprocessing"].update(preprocessing_kwargs)

        # save mha2nnunet_settings
        with open(mha2nnunet_settings_path, "w") as fp:
            json.dump(mha2nnunet_settings, fp, indent=4)
        print(f"Saved updated mha2nnunet settings to {mha2nnunet_settings_path}")


    if nnUNet_dataset_json_path.exists():
        print(f"Found dataset.json at {nnUNet_dataset_json_path}, skipping..")
    else:
        # read preprocessing settings and set the annotation preprocessing function
        with open(mha2nnunet_settings_path) as fp:
            mha2nnunet_settings = json.load(fp)

        if not "options" in mha2nnunet_settings:
            mha2nnunet_settings["options"] = {}
        mha2nnunet_settings["options"]["annotation_preprocess_func"] = preprocess_picai_annotation

        # prepare dataset in nnUNet format
        archive = MHA2nnUNetConverter(
            output_dir=nnUNet_raw_data_path,
            scans_dir=imagesdir,
            annotations_dir=annotations_dir,
            mha2nnunet_settings=mha2nnunet_settings,
        )
        archive.convert()
        archive.create_dataset_json()

    if nnUNet_splits_path.exists():
        print(f"Found cross-validation splits at {nnUNet_splits_path}, skipping..")
    else:
        # save cross-validation splits to disk
        with open(nnUNet_splits_path, "w") as fp:
            json.dump(splits, fp)
        print(f"Saved cross-validation splits to {nnUNet_splits_path}")


if __name__ == "__main__":
    # parse command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--workdir", type=str, default=os.environ.get("workdir", "/workdir"),
                        help="Path to the working directory (default: /workdir, or the environment variable 'workdir')")
    parser.add_argument("--inputdir", type=str, default=os.environ.get("inputdir", "/input"),
                        help="Path to the input dataset (default: /input, or the environment variable 'inputdir')")
    parser.add_argument("--imagesdir", type=str, default="images",
                        help="Path to the images, relative to --inputdir (default: /input/images)")
    parser.add_argument("--labelsdir", type=str, default="picai_labels",
                        help="Path to the labels, relative to --inputdir (root of picai_labels) (default: /input/picai_labels)")
    parser.add_argument("--spacing", type=float, nargs="+", required=False,
                        help="Spacing to preprocess images to. Default: keep as-is.")
    parser.add_argument("--matrix_size", type=int, nargs="+", required=False,
                        help="Matrix size to preprocess images to. Default: keep as-is.")
    parser.add_argument("--preprocessing_kwargs", type=str, required=False,
                        help='Preprocessing kwargs to pass to the MHA2nnUNetConverter. " + \
                            "E.g.: `{"crop_only": true}`. Must be valid json.')
    parser.add_argument("--splits", type=str, default="picai_pub",
                        help="Splits to save for cross-validation. Available: picai_pub, picai_pubpriv.")
    parser.add_argument("--task", type=str, default="Task2203_picai_baseline",
                        help="Task name (default: Task2203_picai_baseline)")
    try:
        args = parser.parse_args()
    except Exception as e:
        print(f"Parsing all arguments failed: {e}")
        print("Retrying with only the known arguments...")
        args, _ = parser.parse_known_args()

    prepare_data_semi_supervised(
        workdir=args.workdir,
        inputdir=args.inputdir,
        imagesdir=args.imagesdir,
        labelsdir=args.labelsdir,
        spacing=args.spacing,
        matrix_size=args.matrix_size,
        preprocessing_kwargs=args.preprocessing_kwargs,
        splits=args.splits,
        task=args.task,
    )
    print("Finished.")



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from pathlib import Path

# read dataset configurations
train_splits = {}
for fold in range(5):
    with open(Path(__file__).parent / f"ds-config-train-fold-{fold}.json") as fp:
        ds_config = json.load(fp)
        train_splits[fold] = ds_config

valid_splits = {}
for fold in range(5):
    with open(Path(__file__).parent / f"ds-config-valid-fold-{fold}.json") as fp:
        ds_config = json.load(fp)
        valid_splits[fold] = ds_config

nnunet_splits = [
    {
        'train': train_splits[fold]["subject_list"],
        'val': valid_splits[fold]["subject_list"],
    }
    for fold in range(5)
]

# expose dataset configurations
__all__ = [
    "train_splits",
    "valid_splits",
    "nnunet_splits",
]


"""
These splits were generated using the following code:

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold

def get_patient_list(subject_list, unique=False):
    patient_list = [s.split("_")[0] for s in subject_list]
    if unique:
        patient_list = list(set(patient_list))
    return patient_list

# settings
folds = range(5)
marksheet_path = "/path/to/pubpriv_training/v1/labels/clinical_information/marksheet.csv"

# read marksheet
marksheet = pd.read_csv(marksheet_path, dtype=str)

# list of studies for training and validation
subject_list_trainval = marksheet.apply(lambda row: f"{row.patient_id}_{row.study_id}", axis=1)
patient_list_trainval = get_patient_list(subject_list_trainval, unique=True)
print(f"Have {len(subject_list_trainval)} studies for training and validation" +
      f" ({len(patient_list_trainval)} patients)")

# sort lists for reproducibility
patient_list_trainval = np.array(sorted(list(patient_list_trainval)))

# placeholders
train_subject_list_map = {fold: [] for fold in folds}
valid_subject_list_map = {fold: [] for fold in folds}

# make splits
kf = KFold(n_splits=len(folds), random_state=576, shuffle=True)
for fold, (_, test_indices) in enumerate(kf.split(patient_list_trainval)):
    patient_ids_valid = patient_list_trainval[test_indices]

    # add to validation set
    valid_subject_list = [s for s in subject_list_trainval
                          if s.split("_")[0] in patient_ids_valid]
    valid_subject_list_map[fold] = valid_subject_list

    # all other studies go in the training set
    train_subject_list = [s for s in subject_list_trainval
                          if not s in valid_subject_list]
    train_subject_list_map[fold] = train_subject_list
"""



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import shutil
from pathlib import Path
from subprocess import check_call

from picai_baseline.splits.picai_nnunet import train_splits, valid_splits
from picai_eval import evaluate_folder
from report_guided_annotation import extract_lesion_candidates

"""
nnUNet baseline for the PI-CAI Challenge
For documentation, please see:
https://github.com/DIAGNijmegen/picai_baseline#nnunet-baseline
"""

# environment settings
if 'inputdir' in os.environ:
    inputdir = Path(os.environ['inputdir'])
else:
    inputdir = Path("/media/pelvis/data/prostate-MRI/picai/public_training")
if 'workdir' in os.environ:
    workdir = Path(os.environ['workdir'])
else:
    workdir = Path("/media/pelvis/projects/joeran/picai/workdir")

# settings
task = "Task2201_picai_baseline"
trainer = "nnUNetTrainerV2_Loss_FL_and_CE_checkpoints"

# paths
mha_archive_dir = inputdir / "images"
annotations_dir = inputdir / "labels/csPCa_lesion_delineations/human_expert/resampled/"
mha2nnunet_settings_path = workdir / "mha2nnunet_settings" / f"{task}.json"
nnUNet_raw_data_path = workdir / "nnUNet_raw_data"
nnUNet_task_dir = nnUNet_raw_data_path / task
nnUNet_dataset_json_path = nnUNet_task_dir / "dataset.json"
nnUNet_splits_path = nnUNet_task_dir / "splits.json"
splits_config_root = Path(__file__).parent / "splits"
nnUNet_trainer_results = workdir / "results/nnUNet/3d_fullres/" / task / "nnUNetTrainerV2_Loss_FL_and_CE_checkpoints__nnUNetPlansv2.1"
nnUNet_gc_algorithm_results = workdir / "picai_nnunet_gc_algorithm/results/nnUNet/3d_fullres/" / task / "nnUNetTrainerV2_Loss_FL_and_CE_checkpoints__nnUNetPlansv2.1"

# train nnUNet (first fold will trigger preprocessing as well)
print("Starting training...")
for fold in range(5):
    # check if training already finished for this fold
    if (workdir / f"results/nnUNet/3d_fullres/{task}/{trainer}__nnUNetPlansv2.1/fold_{fold}/model_final.model").exists():
        print(f"Training already finished for fold {fold}, skipping..")
    else:
        cmd = [
            "nnunet", "plan_train", task,
            workdir.as_posix(),
            "--trainer", trainer,
            "--custom_split", nnUNet_splits_path.as_posix(),
            "--fold", fold
        ]
        check_call([str(c) for c in cmd])
        print(f"Training finished for fold {fold}.")

# perform inference with nnUNet (predict the full train and validation dataset)
print("Starting inference...")
for fold in range(5):
    predictions_dir = nnUNet_task_dir / f"predictions_fold_{fold}"

    # check if inference already finished for this fold
    if predictions_dir.exists():
        print(f"Inference already finished for fold {fold}, skipping..")
    else:
        cmd = [
            "nnunet", "predict", task,
            "--trainer", trainer,
            "--fold", fold,
            "--checkpoint", "model_best",
            "--results", (workdir / "results").as_posix(),
            "--input", (nnUNet_task_dir / "imagesTr").as_posix(),
            "--output", predictions_dir.as_posix(),
            "--store_probability_maps",
        ]
        check_call([str(c) for c in cmd])
        print(f"Inference finished for fold {fold}.")

# perform evaluation
print("Starting evaluation...")
for fold in range(5):
    predictions_dir = nnUNet_task_dir / f"predictions_fold_{fold}"

    # check if evaluation already finished for this fold
    if (workdir / f"predictions_fold_{fold}/metrics.json").exists():
        print(f"Evaluation already finished for fold {fold}, skipping..")
    else:
        metrics = evaluate_folder(
            y_det_dir=predictions_dir,
            y_true_dir=(nnUNet_task_dir / "labelsTr"),
            subject_list=valid_splits[fold],
            y_det_postprocess_func=lambda pred: extract_lesion_candidates(pred)[0],
        )
        metrics.save(predictions_dir / "metrics.json")
        print(f"Evaluation finished for fold {fold}.")

    # optionally, evaluate training performance
    if (workdir / f"predictions_fold_{fold}/metrics-train.json").exists():
        print(f"Evaluation of training performance already finished for fold {fold}, skipping..")
    else:
        metrics = evaluate_folder(
            y_det_dir=predictions_dir,
            y_true_dir=(nnUNet_task_dir / "labelsTr"),
            subject_list=train_splits[fold],
            y_det_postprocess_func=lambda pred: extract_lesion_candidates(pred)[0],
        )
        metrics.save(predictions_dir / "metrics-train.json")
        print(f"Evaluation of training performance finished for fold {fold}.")

# gather resources for inference deployment
for path in [
    "fold_0/model_best.model",
    "fold_0/model_best.model.pkl",
    "fold_1/model_best.model",
    "fold_1/model_best.model.pkl",
    "fold_2/model_best.model",
    "fold_2/model_best.model.pkl",
    "fold_3/model_best.model",
    "fold_3/model_best.model.pkl",
    "fold_4/model_best.model",
    "fold_4/model_best.model.pkl",
    "plans.pkl",
]:
    shutil.copyfile(
        nnUNet_trainer_results / path,
        nnUNet_gc_algorithm_results / path,
    )



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from picai_baseline.unet.training_setup.neural_networks.unets import UNet


def neural_network_for_run(args, device):
    """Select neural network architecture for given run"""

    if args.model_type == 'unet':
        model = UNet(
            spatial_dims=len(args.image_shape),
            in_channels=args.num_channels,
            out_channels=args.num_classes,
            strides=args.model_strides,
            channels=args.model_features
        )
    else:
        raise ValueError(f"Unknown model type: {args.model_type}")

    model = model.to(device)
    print("Loaded Neural Network Arch.:", args.model_type)
    return model



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from collections import OrderedDict
from pathlib import Path

import monai
import numpy as np
import torch
from batchgenerators.dataloading.data_loader import DataLoader
from monai.transforms import Compose, EnsureType
from picai_baseline.unet.training_setup.image_reader import SimpleITKDataset


def default_collate(batch):
    """collate multiple samples into batches, if needed"""

    if isinstance(batch[0], np.ndarray):
        return np.vstack(batch)
    elif isinstance(batch[0], (int, np.int64)):
        return np.array(batch).astype(np.int32)
    elif isinstance(batch[0], (float, np.float32)):
        return np.array(batch).astype(np.float32)
    elif isinstance(batch[0], (np.float64,)):
        return np.array(batch).astype(np.float64)
    elif isinstance(batch[0], (dict, OrderedDict)):
        return {key: default_collate([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], (tuple, list)):
        transposed = zip(*batch)
        return [default_collate(samples) for samples in transposed]
    elif isinstance(batch[0], str):
        return batch
    elif isinstance(batch[0], torch.Tensor):
        return torch.vstack(batch)
    else:
        raise TypeError('unknown type for batch:', type(batch))


class DataLoaderFromDataset(DataLoader):
    """Create dataloader from given dataset"""

    def __init__(self, data, batch_size, num_threads, seed_for_shuffle=1, collate_fn=default_collate,
                 return_incomplete=False, shuffle=True, infinite=False):
        super(DataLoaderFromDataset, self).__init__(data, batch_size, num_threads, seed_for_shuffle,
                                                    return_incomplete=return_incomplete, shuffle=shuffle,
                                                    infinite=infinite)
        self.collate_fn = collate_fn
        self.indices = np.arange(len(data))

    def generate_train_batch(self):

        # randomly select N samples (N = batch size)
        indices = self.get_indices()

        # create dictionary per sample
        batch = [{'data': self._data[i][0].numpy(),
                  'seg': self._data[i][1].numpy()} for i in indices]

        return self.collate_fn(batch)


def prepare_datagens(args, fold_id):
    """Load data sheets --> Create datasets --> Create data loaders"""

    # load datasheets
    with open(Path(args.overviews_dir) / f'PI-CAI_train-fold-{fold_id}.json') as fp:
        train_json = json.load(fp)
    with open(Path(args.overviews_dir) / f'PI-CAI_val-fold-{fold_id}.json') as fp:
        valid_json = json.load(fp)

    # load paths to images and labels
    train_data = [np.array(train_json['image_paths']), np.array(train_json['label_paths'])]
    valid_data = [np.array(valid_json['image_paths']), np.array(valid_json['label_paths'])]

    # use case-level class balance to deduce required train-time class weights
    class_ratio_t = [int(np.sum(train_json['case_label'])), int(len(train_data[0])-np.sum(train_json['case_label']))]
    class_ratio_v = [int(np.sum(valid_json['case_label'])), int(len(valid_data[0])-np.sum(valid_json['case_label']))]
    class_weights = (class_ratio_t / np.sum(class_ratio_t))

    # log dataset definition
    print('Dataset Definition:', "-"*80)
    print(f'Fold Number: {fold_id}')
    print('Data Classes:', list(np.unique(train_json['case_label'])))
    print(f'Train-Time Class Weights: {class_weights}')
    print(f'Training Samples [-:{class_ratio_t[1]};+:{class_ratio_t[0]}]: {len(train_data[1])}')
    print(f'Validation Samples [-:{class_ratio_v[1]};+:{class_ratio_v[0]}]: {len(valid_data[1])}')

    # dummy dataloader for sanity check
    pretx = [EnsureType()]
    check_ds = SimpleITKDataset(image_files=train_data[0][:args.batch_size*2],
                                seg_files=train_data[1][:args.batch_size*2],
                                transform=Compose(pretx),
                                seg_transform=Compose(pretx))
    check_loader = DataLoaderFromDataset(check_ds, batch_size=args.batch_size, num_threads=args.num_threads)
    data_pair = monai.utils.misc.first(check_loader)
    print('DataLoader - Image Shape: ', data_pair['data'].shape)
    print('DataLoader - Label Shape: ', data_pair['seg'].shape)
    print("-"*100)

    assert args.image_shape == list(data_pair['data'].shape[2:])
    assert args.num_channels == data_pair['data'].shape[1]
    assert args.num_classes == len(np.unique(train_json['case_label']))

    # actual dataloaders used at train-time
    train_ds = SimpleITKDataset(image_files=train_data[0], seg_files=train_data[1],
                                transform=Compose(pretx),  seg_transform=Compose(pretx))
    valid_ds = SimpleITKDataset(image_files=valid_data[0], seg_files=valid_data[1],
                                transform=Compose(pretx),  seg_transform=Compose(pretx))
    train_ldr = DataLoaderFromDataset(train_ds, 
        batch_size=args.batch_size, num_threads=args.num_threads, infinite=True, shuffle=True)
    valid_ldr = DataLoaderFromDataset(valid_ds, 
        batch_size=args.batch_size, num_threads=1, infinite=False, shuffle=False)

    return train_ldr, valid_ldr, class_weights.astype(np.float32)



#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from nnunet.training.loss_functions.crossentropy import RobustCrossEntropyLoss
from nnunet.training.network_training.nnUNet_variants.loss_function.nnUNetTrainerV2_focalLoss import \
    FocalLoss
from nnunet.training.network_training.nnUNetTrainerV2 import nnUNetTrainerV2
from torch import nn

# TODO: replace FocalLoss by fixed implemetation (and set smooth=0 in that one?)


class FL_and_CE_loss(nn.Module):
    def __init__(self, fl_kwargs=None, ce_kwargs=None, alpha=0.5, aggregate="sum"):
        super(FL_and_CE_loss, self).__init__()
        if fl_kwargs is None:
            fl_kwargs = {}
        if ce_kwargs is None:
            ce_kwargs = {}

        self.aggregate = aggregate
        self.fl = FocalLoss(apply_nonlin=nn.Softmax(), **fl_kwargs)
        self.ce = RobustCrossEntropyLoss(**ce_kwargs)
        self.alpha = alpha

    def forward(self, net_output, target):
        fl_loss = self.fl(net_output, target)
        ce_loss = self.ce(net_output, target)
        if self.aggregate == "sum":
            result = self.alpha*fl_loss + (1-self.alpha)*ce_loss
        else:
            raise NotImplementedError("nah son")
        return result


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints(nnUNetTrainerV2):
    """
    Set loss to FL + CE and set checkpoints
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)
        self.loss = FL_and_CE_loss(alpha=0.5)
        self.save_latest_only = False


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints2(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints3(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints4(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints5(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints6(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints7(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints8(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints9(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)


class nnUNetTrainerV2_Loss_FL_and_CE_checkpoints10(nnUNetTrainerV2_Loss_FL_and_CE_checkpoints):
    """
    Each run is stored in a folder with the training class name in it. This simply creates a new folder,
    to allow investigating the variability between restarts.
    """

    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)



import setuptools

if __name__ == '__main__':
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setuptools.setup(
        version='0.8.0',
        author_email='Joeran.Bosma@radboudumc.nl',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='https://github.com/DIAGNijmegen/picai_baseline',
        project_urls={
            "Bug Tracker": "https://github.com/DIAGNijmegen/picai_baseline/issues"
        },
        license='Apache 2.0',
        packages=setuptools.find_packages('src', exclude=['tests']),
        package_data={'': [
            'splits/*/*.json',
        ]},
        include_package_data=True,
    )



#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import numpy as np
import torch
import torch.nn as nn
from nnunet.training.network_training.nnUNetTrainerV2 import nnUNetTrainerV2


class FocalLoss(nn.Module):
    """
    copy from: https://github.com/Hsuxu/Loss_ToolBox-PyTorch/blob/master/FocalLoss/FocalLoss.py
    This is a implementation of Focal Loss with smooth label cross entropy supported which is proposed in
    'Focal Loss for Dense Object Detection. (https://arxiv.org/abs/1708.02002)'
        Focal_Loss= -1*alpha*(1-pt)*log(pt)
    :param num_class:
    :param alpha: (tensor) 3D or 4D the scalar factor for this criterion
    :param gamma: (float,double) gamma > 0 reduces the relative loss for well-classified examples (p>0.5) putting more
                    focus on hard misclassified example
    :param smooth: (float,double) smooth value when cross entropy
    :param balance_index: (int) balance class index, should be specific when alpha is float
    :param size_average: (bool, optional) By default, the losses are averaged over each loss element in the batch.
    """

    def __init__(self, apply_nonlin=None, alpha=None, gamma=2, balance_index=0, smooth=1e-5, size_average=True):
        super(FocalLoss, self).__init__()
        self.apply_nonlin = apply_nonlin
        self.alpha = alpha
        self.gamma = gamma
        self.balance_index = balance_index
        self.smooth = smooth
        self.size_average = size_average

        if self.smooth is not None:
            if self.smooth < 0 or self.smooth > 1.0:
                raise ValueError('smooth value should be in [0,1]')

    def forward(self, logit, target):
        if self.apply_nonlin is not None:
            logit = self.apply_nonlin(logit)
        num_class = logit.shape[1]

        if logit.dim() > 2:
            # flatten spatial dimensions N,C,d1,d2 -> N,C,m (m=d1*d2*...)
            logit = logit.view(logit.size(0), logit.size(1), -1)
            logit = logit.permute(0, 2, 1).contiguous()
            logit = logit.view(-1, logit.size(-1))
        target = torch.squeeze(target, 1)
        target = target.view(-1, 1)
        # print(logit.shape, target.shape)

        alpha = self.alpha

        if alpha is None:
            alpha = torch.ones(num_class, 1)
        elif isinstance(alpha, (list, np.ndarray)):
            assert len(alpha) == num_class
            alpha = torch.FloatTensor(alpha).view(num_class, 1)
            alpha = alpha / alpha.sum()
        elif isinstance(alpha, float):
            alpha = torch.ones(num_class, 1)
            alpha = alpha * (1 - self.alpha)
            alpha[self.balance_index] = self.alpha
        else:
            raise TypeError(f'Unsupported alpha type: {type(alpha)}')

        if alpha.device != logit.device:
            alpha = alpha.to(logit.device)

        idx = target.cpu().long()

        one_hot_key = torch.FloatTensor(target.size(0), num_class).zero_()
        one_hot_key = one_hot_key.scatter_(1, idx, 1)
        if one_hot_key.device != logit.device:
            one_hot_key = one_hot_key.to(logit.device)

        if self.smooth:
            one_hot_key = torch.clamp(
                one_hot_key, self.smooth/(num_class-1), 1.0 - self.smooth)
        pt = (one_hot_key * logit).sum(1) + self.smooth
        logpt = pt.log()

        gamma = self.gamma

        alpha = alpha[idx]
        alpha = torch.squeeze(alpha)
        loss = -1 * alpha * torch.pow((1 - pt), gamma) * logpt

        if self.size_average:
            loss = loss.mean()
        else:
            loss = loss.sum()

        return loss


class nnUNetTrainerV2_focalLossAlpha75(nnUNetTrainerV2):
    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)
        print("Setting up FocalLoss(alpha=[0.75, 0.25], apply_nonlin=nn.Softmax())")
        self.loss = FocalLoss(alpha=[0.75, 0.25], apply_nonlin=nn.Softmax())


class nnUNetTrainerV2_focalLossAlpha75_checkpoints(nnUNetTrainerV2_focalLossAlpha75):
    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)
        print("Saving checkpoint every 50th epoch")
        self.save_latest_only = False


class nnUNetTrainerV2_focalLossAlpha75_checkpoints2(nnUNetTrainerV2_focalLossAlpha75_checkpoints):
    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)
        pass  # this is just to get a new Trainer directory


class nnUNetTrainerV2_focalLossAlpha75_checkpoints3(nnUNetTrainerV2_focalLossAlpha75_checkpoints):
    def __init__(self, plans_file, fold, output_folder=None, dataset_directory=None, batch_dice=True, stage=None,
                 unpack_data=True, deterministic=True, fp16=False):
        super().__init__(plans_file, fold, output_folder, dataset_directory, batch_dice, stage, unpack_data,
                         deterministic, fp16)
        pass  # this is just to get a new Trainer directory



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import pickle
from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np
import SimpleITK as sitk
from batchgenerators.utilities.file_and_folder_operations import *
from picai_eval.image_utils import read_prediction

from nnunet.preprocessing.preprocessing import (get_do_separate_z,
                                                get_lowres_axis,
                                                resample_data_or_seg)


# adapted from https://github.com/MIC-DKFZ/nnUNet/blob/master/nnunet/inference/segmentation_export.py
def save_softmax_nifti_from_softmax(segmentation_softmax: Union[str, np.ndarray], out_fname: str,
                                    properties_dict: dict, order: int = 1,
                                    region_class_order: Tuple[Tuple[int]] = None,
                                    seg_postprogess_fn: callable = None, seg_postprocess_args: tuple = None,
                                    resampled_npz_fname: str = None,
                                    non_postprocessed_fname: str = None, force_separate_z: bool = None,
                                    interpolation_order_z: int = 0, verbose: bool = True):
    """
    This is a utility for writing segmentations to nifty and npz. It requires the data to have been preprocessed by
    GenericPreprocessor because it depends on the property dictionary output (dct) to know the geometry of the original
    data. segmentation_softmax does not have to have the same size in pixels as the original data, it will be
    resampled to match that. This is generally useful because the spacings our networks operate on are most of the time
    not the native spacings of the image data.
    If seg_postprogess_fn is not None then seg_postprogess_fnseg_postprogess_fn(segmentation, *seg_postprocess_args)
    will be called before nifty export
    There is a problem with python process communication that prevents us from communicating objects
    larger than 2 GB between processes (basically when the length of the pickle string that will be sent is
    communicated by the multiprocessing.Pipe object then the placeholder (I think) does not allow for long
    enough strings (lol). This could be fixed by changing i to l (for long) but that would require manually
    patching system python code.) We circumvent that problem here by saving softmax_pred to a npy file that will
    then be read (and finally deleted) by the Process. save_segmentation_nifti_from_softmax can take either
    filename or np.ndarray for segmentation_softmax and will handle this automatically
    :param segmentation_softmax:
    :param out_fname:
    :param properties_dict:
    :param order:
    :param region_class_order:
    :param seg_postprogess_fn:
    :param seg_postprocess_args:
    :param resampled_npz_fname:
    :param non_postprocessed_fname:
    :param force_separate_z: if None then we dynamically decide how to resample along z, if True/False then always
    /never resample along z separately. Do not touch unless you know what you are doing
    :param interpolation_order_z: if separate z resampling is done then this is the order for resampling in z
    :param verbose:
    :return:
    """
    if verbose: print("force_separate_z:", force_separate_z, "interpolation order:", order)

    if isinstance(segmentation_softmax, str):
        assert os.path.isfile(segmentation_softmax), "If isinstance(segmentation_softmax, str) then " \
                                                     "isfile(segmentation_softmax) must be True"
        # del_file = deepcopy(segmentation_softmax)  # <- skipped this
        if segmentation_softmax.endswith('.npy'):
            segmentation_softmax = np.load(segmentation_softmax)
        elif segmentation_softmax.endswith('.npz'):
            segmentation_softmax = np.load(segmentation_softmax)['softmax']
        # os.remove(del_file)  # <- skipped this

    # first resample, then put result into bbox of cropping, then save
    current_shape = segmentation_softmax.shape
    shape_original_after_cropping = properties_dict.get('size_after_cropping')
    shape_original_before_cropping = properties_dict.get('original_size_of_raw_data')
    # current_spacing = dct.get('spacing_after_resampling')
    # original_spacing = dct.get('original_spacing')

    if np.any([i != j for i, j in zip(np.array(current_shape[1:]), np.array(shape_original_after_cropping))]):
        if force_separate_z is None:
            if get_do_separate_z(properties_dict.get('original_spacing')):
                do_separate_z = True
                lowres_axis = get_lowres_axis(properties_dict.get('original_spacing'))
            elif get_do_separate_z(properties_dict.get('spacing_after_resampling')):
                do_separate_z = True
                lowres_axis = get_lowres_axis(properties_dict.get('spacing_after_resampling'))
            else:
                do_separate_z = False
                lowres_axis = None
        else:
            do_separate_z = force_separate_z
            if do_separate_z:
                lowres_axis = get_lowres_axis(properties_dict.get('original_spacing'))
            else:
                lowres_axis = None

        if lowres_axis is not None and len(lowres_axis) != 1:
            # this happens for spacings like (0.24, 1.25, 1.25) for example. In that case we do not want to resample
            # separately in the out of plane axis
            do_separate_z = False

        if verbose: print("separate z:", do_separate_z, "lowres axis", lowres_axis)
        seg_old_spacing = resample_data_or_seg(segmentation_softmax, shape_original_after_cropping, is_seg=False,
                                               axis=lowres_axis, order=order, do_separate_z=do_separate_z,
                                               order_z=interpolation_order_z)
        # seg_old_spacing = resize_softmax_output(segmentation_softmax, shape_original_after_cropping, order=order)
    else:
        if verbose: print("no resampling necessary")
        seg_old_spacing = segmentation_softmax

    if resampled_npz_fname is not None:
        np.savez_compressed(resampled_npz_fname, softmax=seg_old_spacing.astype(np.float16))
        # this is needed for ensembling if the nonlinearity is sigmoid
        if region_class_order is not None:
            properties_dict['regions_class_order'] = region_class_order
        save_pickle(properties_dict, resampled_npz_fname[:-4] + ".pkl")

    # skip this, we don't want to threshold/argmax softmax output
    # if region_class_order is None:
    #     seg_old_spacing = seg_old_spacing.argmax(0)
    # else:
    #     seg_old_spacing_final = np.zeros(seg_old_spacing.shape[1:])
    #     for i, c in enumerate(region_class_order):
    #         seg_old_spacing_final[seg_old_spacing[i] > 0.5] = c
    #     seg_old_spacing = seg_old_spacing_final
    
    if seg_old_spacing.shape[0] == 2:
        # binary softmax, save foreground only
        seg_old_spacing = seg_old_spacing[1]
    else:
        raise NotImplementedError("only export of binary softmax is implemented")

    bbox = properties_dict.get('crop_bbox')

    if bbox is not None:
        seg_old_size = np.zeros(shape_original_before_cropping, dtype=np.float32)  # <- changed to float32
        for c in range(3):
            bbox[c][1] = np.min((bbox[c][0] + seg_old_spacing.shape[c], shape_original_before_cropping[c]))
        seg_old_size[
            bbox[0][0]:bbox[0][1],
            bbox[1][0]:bbox[1][1],
            bbox[2][0]:bbox[2][1]
        ] = seg_old_spacing
    else:
        seg_old_size = seg_old_spacing

    if seg_postprogess_fn is not None:
        seg_old_size_postprocessed = seg_postprogess_fn(np.copy(seg_old_size), *seg_postprocess_args)
    else:
        seg_old_size_postprocessed = seg_old_size

    seg_resized_itk = sitk.GetImageFromArray(seg_old_size_postprocessed.astype(np.float32))  # <- changed to float32
    seg_resized_itk.SetSpacing(properties_dict['itk_spacing'])
    seg_resized_itk.SetOrigin(properties_dict['itk_origin'])
    seg_resized_itk.SetDirection(properties_dict['itk_direction'])
    sitk.WriteImage(seg_resized_itk, out_fname)

    if (non_postprocessed_fname is not None) and (seg_postprogess_fn is not None):
        seg_resized_itk = sitk.GetImageFromArray(seg_old_size.astype(np.uint8))
        seg_resized_itk.SetSpacing(properties_dict['itk_spacing'])
        seg_resized_itk.SetOrigin(properties_dict['itk_origin'])
        seg_resized_itk.SetDirection(properties_dict['itk_direction'])
        sitk.WriteImage(seg_resized_itk, non_postprocessed_fname)


def convert_cropped_npz_to_original_nifty(
    npz_path: Union[Path, str],
    pkl_path: Optional[Union[Path, str]] = None,
    dst_path: Optional[Union[Path, str]] = None,
    overwrite: bool = False,
) -> None:
    """
    Convert nnUNet's softmax prediction (stored as .npz) to
    the scan's original extent (before cropping).

    Parameters
    ----------
    - npz_path: path to prediction as .npz file
    - pkl_path: path to metadata about nnU-Net preprocessing
        default: npz_path with .pkl instead of .npz
    - dst_path: path to save converted softmax to
        default: npz_path with _softmax.nii.gz instead of .npz
    - overwrite: whether to overwrite an existing softmax prediction
    """
    npz_path = Path(npz_path)
    if pkl_path is None:
        pkl_path = npz_path.with_name(f"{npz_path.stem}.pkl")
    if dst_path is None:
        dst_path = npz_path.with_name(f"{npz_path.stem}_softmax.nii.gz")
    else:
        dst_path = Path(dst_path)
    if not overwrite and dst_path.exists():
        return

    # read prediction and preprocessing info
    # read predictions from all folds
    pred = read_prediction(npz_path)

    # convert to nnUNet's internal softmax format
    pred = np.array([1-pred, pred])

    # read physical properties of current case
    with open(pkl_path, "rb") as fp:
        properties = pickle.load(fp)

    # let nnUNet resample to original physical space
    save_softmax_nifti_from_softmax(
        segmentation_softmax=pred,
        out_fname=str(dst_path),
        properties_dict=properties,
    )



#!/opt/conda/bin/python

import argparse
import functools
import os
import pickle
import re
import subprocess
import sys
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path

import numpy as np
from carbontracker.tracker import CarbonTracker
from nnunet.utilities import shutil_sol
from nnunet.utilities.io import (checksum, path_exists, read_json,
                                 refresh_file_list, write_json)
from picai_prep.data_utils import atomic_file_copy

PLANS = 'nnUNetPlansv2.1'

# add a flush to each print statement to ensure the stuff gets logged on SOL
print = functools.partial(print, flush=True)


class CustomizedCarbonTracker:
    def __init__(self, logdir, enabled=True):
        if enabled:
            self.tracker = CarbonTracker(epochs=1, ignore_errors=True, devices_by_pid=False, log_dir=str(logdir),
                                         verbose=2)
        else:
            self.tracker = None

    def __enter__(self):
        if self.enabled:
            self.tracker.epoch_start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enabled:
            self.tracker.epoch_end()
            self.tracker.stop()

    @property
    def enabled(self):
        return self.tracker is not None


def get_task_id(task_name):
    return re.match('Task([0-9]+)', task_name).group(1)


def print_split_per_fold(split_file, fold=None):
    try:
        with split_file.open('rb') as pkl:
            splits = pickle.load(pkl)
    except FileNotFoundError:
        print('Split file not found')
    else:
        for i, split in enumerate(splits):
            if fold not in (None, i):
                continue

            print(f'Fold #{i}')
            print('> Training')
            for caseid in sorted(split['train']):
                print(f'>> {caseid}')
            print('> Validation')
            for caseid in sorted(split['val']):
                print(f'>> {caseid}')

            if i + 1 < len(splits):
                print('-' * 25)


def prepare(argv):
    # Convert images and masks into expected format
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('data', type=str)
    parser.add_argument('--images', type=str, required=True)
    parser.add_argument('--masks', type=str, required=True)
    parser.add_argument('--modality', type=str, default='CT')
    parser.add_argument('--labels', type=str, nargs='*', default=['background', 'foreground'])
    parser.add_argument('--license', type=str, default='')
    parser.add_argument('--release', type=str, default='1.0')
    args = parser.parse_args(argv)

    # Check if task already exists
    print('[#] Creating directory structure')

    datadir = Path(args.data)
    taskdir = datadir / 'nnUNet_raw_data' / args.task

    try:
        taskdir.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        print(f'Destination "{taskdir}" already exists')
        return

    # Determine image source and destination directories
    image_srcdir = Path(args.images)
    if '*' in image_srcdir.name:
        image_glob_pattern = image_srcdir.name
        image_srcdir = image_srcdir.parent
    else:
        image_glob_pattern = '*.mha'

    image_dstdir = taskdir / 'imagesTr'
    image_dstdir.mkdir()

    # Prepare mask source and destianation directories
    mask_srcdir = Path(args.masks)
    mask_dstdir = taskdir / 'labelsTr'
    mask_dstdir.mkdir()

    # Copy / convert images and masks
    print('[#] Converting images and masks')
    training = []
    for image_srcfile in sorted(image_srcdir.glob(image_glob_pattern)):
        if image_srcfile.name.startswith('.'):
            # For some reason, hidden files are sometimes included and Mac users run
            # into problems then because they have plenty of ._ metadata files.
            continue

        if image_srcfile.name.endswith('.nii.gz'):
            caseid = image_srcfile.name[:-7]
            ext = 'nii.gz'
        else:
            caseid = image_srcfile.stem
            ext = image_srcfile.suffix[1:]

        if caseid.endswith('_0000'):
            caseid = caseid[:-5]

        try:
            mask_srcfile = mask_srcdir / f'{caseid}.{ext}'
            if not mask_srcfile.exists():
                mask_srcfile = next(mask_srcdir.glob(f'{caseid}_*.{ext}'))
        except StopIteration:
            print(f'Missing mask for case "{caseid}"')
            return

        image_dstfile = image_dstdir / f'{caseid}_0000.nii.gz'
        print(f'{image_srcfile.name} -> {image_dstfile.name}')
        atomic_file_copy(image_srcfile, image_dstfile)

        mask_dstfile = mask_dstdir / f'{caseid}.nii.gz'
        atomic_file_copy(mask_srcfile, mask_dstfile)

        training.append({
            'image': f'./imagesTr/{caseid}.nii.gz',
            'label': f'./labelsTr/{caseid}.nii.gz'
        })

    # Create metadata file
    print('[#] Writing metadata to dataset.json')

    name = args.task.split('_', 1)[1]
    labels = OrderedDict([(str(i), label) for i, label in enumerate(args.labels)])
    metadata = OrderedDict([
        ('name', name),
        ('description', f'{name}, reformatted for nnU-net'),
        ('tensorImageSize', '3D'),
        ('licence', args.license),
        ('release', args.release),
        ('modality', {'0': args.modality}),
        ('labels', labels),
        ('numTraining', len(training)),
        ('numTest', 0),
        ('training', training),
        ('test', [])
    ])
    write_json(taskdir / 'dataset.json', metadata, make_dirs=False)


def plan_train(argv):
    # Plan experiment, then train network
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('data', type=str)
    parser.add_argument('--results', type=str, required=False)
    parser.add_argument('--network', type=str, default='3d_fullres')
    parser.add_argument('--trainer', type=str, default='nnUNetTrainerV2')
    parser.add_argument('--trainer_kwargs', required=False, default="{}",
                        help="Use a dictionary in string format to specify keyword arguments. This will get"
                             " parsed into a dictionary, the values get correctly parsed to the data format"
                             " and passed to the trainer. Example (backslash included): \n"
                             r"--trainer_kwargs='{\"class_weights\":[0,2.00990337,1.42540704,2.13387239,0.85529504,0.592059,0.30040984,8.26874351],"
                             r"\"weight_dc\":0.3,\"weight_ce\":0.7}'")
    parser.add_argument('--kwargs', type=str, required=False, default=None,
                        help="Specify additional arguments for nnUNet_train. Example (backslash included): \n"
                             r"--kwargs='--disable_postprocessing_on_folds'")
    parser.add_argument('--fold', type=str, default='0')
    parser.add_argument('--custom_split', type=str, help='Path to a JSON file with a custom data split into five folds')
    parser.add_argument('--plan_only', action='store_true', help='Run the planning step, but not the training step')
    parser.add_argument('--validation_only', action='store_true',
                        help='Do no run network training, only the final validation step')
    parser.add_argument('--ensembling', action='store_true',
                        help='Export probability maps for ensembling during the final validation')
    parser.add_argument('--use_compressed_data', action='store_true',
                        help='Disable unpacking of compressed training data, use with caution')
    parser.add_argument('--plan_2d', action='store_true', help='Enable planning of 2D experiments')
    parser.add_argument('--dont_plan_3d', action='store_true', help='Disable planning of 3D experiments')
    parser.add_argument('--carbontracker', action='store_true', help='Enables tracking of energy consumption')
    parser.add_argument('--pretrained_weights', type=str, required=False, default=None)
    args = parser.parse_args(argv)

    # aid type hinting
    args.task = str(args.task)

    # Set environment variables
    datadir = Path(args.data)
    prepdir = Path(os.environ.get('prepdir', '/home/user/data'))

    splits_file = prepdir / args.task / 'splits_final.pkl'

    os.environ['nnUNet_raw_data_base'] = str(datadir)
    os.environ['nnUNet_preprocessed'] = str(prepdir)
    os.environ['RESULTS_FOLDER'] = args.results if args.results else str(datadir / 'results')

    # Start carbontracker
    with CustomizedCarbonTracker(prepdir / 'carbontracker', enabled=args.carbontracker):
        # Check if plans and preprocessed data are available
        taskid = get_task_id(args.task)
        taskdir = datadir / 'nnUNet_preprocessed' / args.task

        if path_exists(taskdir):
            if args.custom_split:
                remote_splits_file = taskdir / 'splits_final.json'
                if not remote_splits_file.exists() or checksum(remote_splits_file) != checksum(args.custom_split):
                    print(f"[#] Found plans and preprocessed data for {args.task}"
                          " - but you also provided a custom split which is different"
                          " from the present split, this is not permitted")
                    return

            if args.plan_only:
                print(f'[#] Found plans and preprocessed data for {args.task} - nothing to do')
            else:
                print(f'[#] Found plans and preprocessed data for {args.task} - copying to compute node')
                if not os.path.exists(prepdir / args.task):
                    prepdir.mkdir(parents=True, exist_ok=True)
                    shutil_sol.copytree(taskdir, prepdir / args.task)
                print(f'[#] Found plans and preprocessed data for {args.task} - copied to compute node')
        else:
            # Plans and data not available yet, run preprocessing
            print('[#] Creating plans and preprocessing data')
            cmd = [
                'nnUNet_plan_and_preprocess',
                '-t', taskid,
                '-tl', os.environ.get("nnUNet_tl", '1'), '-tf', os.environ.get("nnUNet_tf", '1'),
                '--verify_dataset_integrity'
            ]
            if not args.plan_2d and '2d' not in args.network:
                cmd.extend(['--planner2d', 'None'])  # disable 2D planning to speed up the preprocessing phase
            if args.dont_plan_3d and '3d' not in args.network:
                cmd.extend(['--planner3d', 'None'])
            if args.pretrained_weights is not None:
                cmd.extend(['-pretrained_weights', args.pretrained_weights])
            subprocess.check_call(cmd)

            # Use a custom data split?
            if args.custom_split:
                splits = []
                for split in read_json(args.custom_split):
                    splits.append(OrderedDict([
                        ('train', np.array(split['train'])),
                        ('val', np.array(split['val']))
                    ]))

                splits_file.parent.mkdir(parents=True, exist_ok=True)
                with splits_file.open('wb') as fp:
                    pickle.dump(splits, fp)
                shutil_sol.copyfile(args.custom_split, splits_file.with_suffix('.json'))

            if (prepdir / args.task).absolute() != taskdir.absolute():
                # Copy preprocessed data to storage server
                print('[#] Copying plans and preprocessed data from compute node to storage server')
                taskdir.parent.mkdir(parents=True, exist_ok=True)
                shutil_sol.copytree(prepdir / args.task, taskdir)

        if args.plan_only:
            return

        # Run training
        cmd = [
            'nnUNet_train',
            args.network,
            args.trainer,
            taskid,
            args.fold
        ]

        fold_name = 'all' if args.fold == 'all' else f'fold_{args.fold}'
        outdir = Path(
            os.environ['RESULTS_FOLDER']) / 'nnUNet' / args.network / args.task / f'{args.trainer}__{PLANS}' / fold_name

        if args.validation_only:
            print('[#] Running validation step only')
            cmd.append('--validation_only')
        elif path_exists(outdir) and any(outdir.glob("*.model")):
            print('[#] Resuming network training')
            cmd.append('-c')
        else:
            print('[#] Starting network training')

        if args.trainer_kwargs:
            cmd.append('--trainer_kwargs=%s' % args.trainer_kwargs)
        if args.use_compressed_data:
            cmd.append('--use_compressed_data')
        if args.ensembling:
            cmd.append('--npz')
        if args.kwargs is not None:
            cmd.extend(args.kwargs.split(" "))
        print(f'[#] Running {" ".join(cmd)}')

        subprocess.check_call(cmd)

        # Copy split file since that is for sure available now (nnUNet_train has created
        # it if the file did not exist already - unless training with "all", so still check)
        if splits_file.exists() and splits_file.parent.absolute() != taskdir.absolute():
            shutil_sol.copyfile(splits_file, taskdir)


def reveal_split(argv):
    # Print out the 5-fold cross validation split
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('data', type=str)
    args = parser.parse_args(argv)

    # Locate split file
    datadir = Path(args.data)
    split_file = datadir / 'nnUNet_preprocessed' / args.task / 'splits_final.pkl'

    print_split_per_fold(split_file)


def find_best_configuration(argv):
    # Find best configuration by analyzing cross-validation results
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('data', type=str)
    parser.add_argument('--networks', type=str, nargs='*', default=['3d_fullres'])
    parser.add_argument('--trainer', type=str, default='nnUNetTrainerV2')
    args = parser.parse_args(argv)

    # Set environment variables
    datadir = Path(args.data)
    prepdir = datadir / 'nnUNet_preprocessed'
    os.environ['nnUNet_preprocessed'] = str(prepdir)
    os.environ['RESULTS_FOLDER'] = str(datadir / 'results')

    # Prepare output directory to prevent crashes
    print('[#] Preparing output directory')
    cmdir = datadir / 'results' / 'nnUNet' / 'ensembles' / args.task
    cmdir.mkdir(parents=True, exist_ok=True)

    # Run scripts
    for network in args.networks:
        print(f'[#] Consolidating folds for {network} network')
        subprocess.check_call([
            'nnUNet_determine_postprocessing',
            '-m', network,
            '-t', get_task_id(args.task),
            '-tr', str(args.trainer)
        ])

    if len(args.networks) > 1:
        print('[#] Finding best configuration')
        refresh_file_list(prepdir / args.task / 'gt_segmentations')
        subprocess.check_call([
            'nnUNet_find_best_configuration',
            '-m', *args.networks,
            '-t', get_task_id(args.task),
            '-tr', str(args.trainer)
        ])


def _predict(args):
    # Set environment variables
    os.environ['RESULTS_FOLDER'] = args.results

    # Prepare output directory to prevent crashes
    outdir = Path(args.output).absolute()
    outdir.mkdir(parents=True, exist_ok=True)

    # Run prediction script
    cmd = [
        'nnUNet_predict',
        '-t', args.task,
        '-i', args.input,
        '-o', args.output,
        '-m', args.network,
        '-tr', args.trainer,
        '--num_threads_preprocessing', '2',
        '--num_threads_nifti_save', '1'
    ]

    if args.folds:
        cmd.append('-f')
        cmd.extend(args.folds.split(','))

    if args.checkpoint:
        cmd.append('-chk')
        cmd.append(args.checkpoint)

    if args.store_probability_maps:
        cmd.append('--save_npz')

    if args.disable_augmentation:
        cmd.append('--disable_tta')

    if args.disable_patch_overlap:
        cmd.extend(['--step_size', '1'])

    subprocess.check_call(cmd)


def predict(argv):
    # Use trained network to generate segmentation masks
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('--input', type=str, default='/input')
    parser.add_argument('--output', type=str, default='/output')
    parser.add_argument('--results', type=str, required=True)  # Path to training results folder with model weights etc
    parser.add_argument('--network', type=str, default='3d_fullres')
    parser.add_argument('--trainer', type=str, default='nnUNetTrainerV2')
    parser.add_argument('--folds', type=str, required=False)
    parser.add_argument('--checkpoint', type=str,
                        required=False)  # Checkpoint to load, defaults to "model_final_checkpoint"
    parser.add_argument('--store_probability_maps', action='store_true')
    parser.add_argument('--disable_augmentation', action='store_true')
    parser.add_argument('--disable_patch_overlap', action='store_true')
    args = parser.parse_args(argv)

    _predict(args)


def ensemble(argv):
    # Run inference and ensemble predictions
    parser = argparse.ArgumentParser()
    parser.add_argument('task', type=str)
    parser.add_argument('--input', type=str, default='/input')
    parser.add_argument('--output', type=str, default='/output')
    parser.add_argument('--results', type=str, required=True)  # Path to training results folder with model weights etc
    parser.add_argument('--networks', type=str, nargs='*', default=['3d_fullres'])
    parser.add_argument('--trainers', type=str, nargs='*', default=['nnUNetTrainerV2'])
    parser.add_argument('--folds', type=str, required=False)
    parser.add_argument('--checkpoint', type=str,
                        required=False)  # Checkpoint to load, defaults to "model_final_checkpoint"
    parser.add_argument('--disable_augmentation', action='store_true')
    parser.add_argument('--disable_patch_overlap', action='store_true')
    args = parser.parse_args(argv)

    # Run inference for all networks in the ensemble
    output_dirs = []
    ensemble_name_fragments = []
    for i, network in enumerate(args.networks):
        print(f'[#] Running inference for {network} network')

        args_predict = deepcopy(args)
        args_predict.store_probability_maps = True

        args_predict.network = network
        del args_predict.networks

        args_predict.trainer = args.trainers[i] if len(args.trainers) > i else args.trainers[-1]
        del args_predict.trainers

        output_dir = Path(args.output) / network
        output_dirs.append(output_dir)
        args_predict.output = str(output_dir)

        ensemble_name_fragments.append(f'{args_predict.network}__{args_predict.trainer}__{PLANS}')
        _predict(args_predict)

    # Combine results
    print('[#] Ensembling results')
    ensemble_name = 'ensemble_' + '--'.join(ensemble_name_fragments)
    output_dir = Path(args.output) / ensemble_name

    cmd = [
        'nnUNet_ensemble',
        '-f', *[str(f) for f in output_dirs],
        '-o', str(output_dir)
    ]

    pp_file = Path(args.results) / 'nnUNet' / 'ensembles' / args.task / ensemble_name / 'postprocessing.json'
    if path_exists(pp_file):
        cmd.extend(['-pp', str(pp_file)])

    subprocess.check_call(cmd)


def evaluate(argv):
    # Run evaluation on the test set
    parser = argparse.ArgumentParser()
    parser.add_argument('--ground_truth', type=str, required=True)
    parser.add_argument('--prediction', type=str, required=True)
    parser.add_argument('--labels', type=str, nargs='+', required=True)
    args = parser.parse_args(argv)

    # Check if the specified folders exist
    ground_truth_dir = Path(args.ground_truth)
    if not path_exists(ground_truth_dir):
        print('Folder with ground truth annotations does not exist')
        return

    prediction_dir = Path(args.prediction)
    if not path_exists(prediction_dir):
        print('Folder with ground truth annotations does not exist')
        return

    # Parse labels
    range_pattern = re.compile('[0-9]+-[0-9]+')
    if len(args.labels) == 1 and range_pattern.fullmatch(args.labels[0]):
        r = tuple(map(int, args.labels[0].split('-')))
        labels = list(map(str, range(r[0], r[1] + 1)))
    else:
        labels = args.labels

    # Run nnUNet script
    print('[#] Evaluating test set results')
    subprocess.check_call([
        'nnUNet_evaluate_folder',
        '-ref', str(ground_truth_dir),
        '-pred', str(prediction_dir),
        '-l', *labels
    ])

    # Read results file
    results_file = prediction_dir / 'summary.json'
    try:
        results = read_json(results_file)
    except IOError:
        print('Evaluation failed')
        return

    print('Average Dice scores across all cases:')
    for label, metrics in sorted(results['results']['mean'].items(), key=lambda item: int(item[0])):
        print(f' > {label}: {metrics["Dice"]}')


def checkout(argv):
    # Switch to a specific branch of the nnU-Net repository?
    parser = argparse.ArgumentParser()
    parser.add_argument('--checkout', type=str, default='')
    args, unknown = parser.parse_known_args(argv)

    if args.checkout:
        subprocess.check_call([
            'git', '-C', '/home/user/nnunet',
            'fetch'
        ])
        subprocess.check_call([
            'git', '-C', '/home/user/nnunet',
            'checkout', args.checkout
        ])

    return unknown


if __name__ == '__main__':
    # Very first argument determines action
    actions = {
        'prepare': prepare,
        'plan_train': plan_train,
        'reveal_split': reveal_split,
        'find_best_configuration': find_best_configuration,
        'predict': predict,
        'ensemble': ensemble,
        'evaluate': evaluate
    }

    try:
        action = actions[sys.argv[1]]
        argv = checkout(sys.argv[2:])
    except (IndexError, KeyError):
        print('Usage: nnunet ' + '/'.join(actions.keys()) + ' ...')
    else:
        action(argv)



#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import time
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from picai_baseline.unet.training_setup.poly_lr import poly_lr
from picai_eval import evaluate
from report_guided_annotation import extract_lesion_candidates
from scipy.ndimage import gaussian_filter


def resume_or_restart_training(model, optimizer, device, args, fold_id):
    """Resume/restart training, based on whether checkpoint exists"""

    weights_file = Path(args.weights_dir) / f"{args.model_type}_F{fold_id}.pt"
    metrics_file = Path(args.weights_dir) / f"{args.model_type}_F{fold_id}_metrics.xlsx"

    if bool(args.resume_training) and weights_file.is_file():
        print("Loading Weights From:", weights_file)
        checkpoint = torch.load(weights_file)

        # load weights and optimizer state
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        model.to(device)

        # load train-time metrics from interrupted run
        tracking_metrics = {}
        if metrics_file.is_file():

            saved_metrics = pd.read_excel(metrics_file, engine='openpyxl')
            all_epochs = (saved_metrics['epoch'].values).tolist()
            all_valid_metrics_auroc = (saved_metrics['valid_auroc'].values).tolist()
            all_valid_metrics_ap = (saved_metrics['valid_ap'].values).tolist()
            all_valid_metrics_ranking = (saved_metrics['valid_ranking'].values).tolist()

            tracking_metrics = {
                'fold_id':                   fold_id,
                'start_epoch':               checkpoint['epoch'] + 1,  # resume at next epoch
                'all_epochs':                all_epochs,
                'all_train_loss':           (saved_metrics['train_loss'].values).tolist(),
                'all_valid_metrics_auroc':   all_valid_metrics_auroc,
                'all_valid_metrics_ap':      all_valid_metrics_ap,
                'all_valid_metrics_ranking': all_valid_metrics_ranking,
                'best_metric':               np.max(all_valid_metrics_ranking),
                'best_metric_epoch':         all_epochs[all_valid_metrics_ranking.index(
                    np.max(all_valid_metrics_ranking))]}

            print('Previous Record of Metrics Loaded:', metrics_file)
        else:
            print('Previous Record of Metrics Not Found:', metrics_file)

            tracking_metrics = {
                'fold_id':                    fold_id,
                'start_epoch':                checkpoint['epoch'],
                'all_epochs':                 [],
                'all_train_loss':             [],
                'all_valid_metrics_auroc':    [],
                'all_valid_metrics_ap':       [],
                'all_valid_metrics_ranking':  [],
                'best_metric': -1,
                'best_metric_epoch': -1}

        print("Resume Training: Epoch",  tracking_metrics['start_epoch']+1)
        print("Best Validation Metric:", tracking_metrics['best_metric'],
              "@ Epoch", tracking_metrics['best_metric_epoch'])
    else:
        tracking_metrics = {
            'fold_id':                    fold_id,
            'start_epoch':                0,
            'all_epochs':                 [],
            'all_train_loss':             [],
            'all_valid_metrics_auroc':    [],
            'all_valid_metrics_ap':       [],
            'all_valid_metrics_ranking':  [],
            'best_metric': -1,
            'best_metric_epoch': -1}

        print("Start Training: Epoch", tracking_metrics['start_epoch']+1)

    return model, optimizer, tracking_metrics


def optimize_model(model, optimizer, loss_func, train_gen, args, tracking_metrics, device, writer):
    """Optimize model x N training steps per epoch + update learning rate"""

    train_loss, step = 0,  0
    start_time = time.time()
    epoch = tracking_metrics['epoch']

    # for each mini-batch or optimization step
    for batch_data in train_gen:
        step += 1
        try:
            inputs = batch_data['data'].to(device)
            labels = batch_data['seg'].to(device)
        except Exception:
            inputs = torch.from_numpy(batch_data['data']).to(device)
            labels = torch.from_numpy(batch_data['seg']).to(device)
        outputs = model(inputs)
        loss = loss_func(outputs, labels[:, 0, ...].long())
        train_loss += loss.item()

        # backpropagate + optimize
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # define each training epoch == 100 steps (note: nnU-Net uses 250 steps)
        if step >= 100: 
            break

    # update learning rate
    updated_lr = poly_lr(epoch+1, args.num_epochs, args.base_lr, 0.95)
    optimizer.param_groups[0]['lr'] = updated_lr
    print("Learning Rate Updated! New Value: "+str(np.round(updated_lr, 10)), flush=True)

    # track training metrics
    train_loss /= step
    tracking_metrics['train_loss'] = train_loss
    writer.add_scalar("train_loss", train_loss, epoch+1)
    print("-" * 100)
    print(f"Epoch {epoch + 1}/{args.num_epochs} (Train. Loss: {train_loss:.4f}; \
        Time: {int(time.time()-start_time)}sec; Steps Completed: {step})", flush=True)

    return model, optimizer, train_gen, tracking_metrics, writer


def validate_model(model, optimizer, valid_gen, args, tracking_metrics, device, writer):
    """Validate model per N epoch + export model weights"""

    all_valid_preds, all_valid_labels = [], []
    epoch, f = tracking_metrics['epoch'], tracking_metrics['fold_id']

    # for each validation sample
    for valid_data in valid_gen:

        try:
            valid_images = valid_data['data'].to(device)
            valid_labels = valid_data['seg']
        except Exception:
            valid_images = torch.from_numpy(valid_data['data']).to(device)
            valid_labels = torch.from_numpy(valid_data['seg'])

        # test-time augmentation
        valid_images = [valid_images, torch.flip(valid_images, [4]).to(device)]

        # aggregate all validation predictions
        # gaussian blur to counteract checkerboard artifacts in
        # predictions from the use of transposed conv. in the U-Net
        preds = [
            torch.sigmoid(model(x))[:, 1, ...].detach().cpu().numpy()
            for x in valid_images
        ]

        # revert horizontally flipped tta image
        preds[1] = np.flip(preds[1], [3])

        # gaussian blur to counteract checkerboard artifacts in
        # predictions from the use of transposed conv. in the U-Net
        all_valid_preds += [
            np.mean([
                gaussian_filter(x, sigma=1.5)
                for x in preds
            ], axis=0)
        ]
        all_valid_labels += [valid_labels.numpy()[:, 0, ...]]

    # track validation metrics
    valid_metrics = evaluate(y_det=iter(np.concatenate([x for x in np.array(all_valid_preds)], axis=0)),
                             y_true=iter(np.concatenate([x for x in np.array(all_valid_labels)], axis=0)),
                             y_det_postprocess_func=lambda pred: extract_lesion_candidates(pred)[0])

    num_pos = int(np.sum([np.max(y) for y in np.concatenate(
        [x for x in np.array(all_valid_labels)], axis=0)]))
    num_neg = int(len(np.concatenate([x for x in
                                      np.array(all_valid_labels)], axis=0)) - num_pos)

    tracking_metrics['all_epochs'].append(epoch+1)
    tracking_metrics['all_train_loss'].append(tracking_metrics['train_loss'])
    tracking_metrics['all_valid_metrics_auroc'].append(valid_metrics.auroc)
    tracking_metrics['all_valid_metrics_ap'].append(valid_metrics.AP)
    tracking_metrics['all_valid_metrics_ranking'].append(valid_metrics.score)

    # export train-time + validation metrics as .xlsx sheet
    metricsData = pd.DataFrame(list(zip(tracking_metrics['all_epochs'],
                                        tracking_metrics['all_train_loss'],
                                        tracking_metrics['all_valid_metrics_auroc'],
                                        tracking_metrics['all_valid_metrics_ap'],
                                        tracking_metrics['all_valid_metrics_ranking'])),
                               columns=['epoch', 'train_loss', 'valid_auroc', 'valid_ap', 'valid_ranking'])

    # create target folder and save exports sheet
    os.makedirs(args.weights_dir, exist_ok=True)

    metrics_file = Path(args.weights_dir) / f"{args.model_type}_F{f}_metrics.xlsx"
    metricsData.to_excel(metrics_file, encoding='utf-8', index=False)

    writer.add_scalar("valid_auroc",   valid_metrics.auroc, epoch+1)
    writer.add_scalar("valid_ap",      valid_metrics.AP,    epoch+1)
    writer.add_scalar("valid_ranking", valid_metrics.score, epoch+1)

    print(f"Valid. Performance [Benign or Indolent PCa (n={num_neg}) \
        vs. csPCa (n={num_pos})]:\nRanking Score = {valid_metrics.score:.3f},\
        AP = {valid_metrics.AP:.3f}, AUROC = {valid_metrics.auroc:.3f}", flush=True)

    # store model checkpoint if validation metric improves
    if valid_metrics.score > tracking_metrics['best_metric']:
        tracking_metrics['best_metric'] = valid_metrics.score
        tracking_metrics['best_metric_epoch'] = epoch + 1
        if bool(args.export_best_model):
            weights_file = Path(args.weights_dir) / f"{args.model_type}_F{f}.pt"

            print("Validation Ranking Score Improved! Saving New Best Model", flush=True)
            torch.save({
                'epoch': epoch,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict()
            }, weights_file)
    return model, optimizer, valid_gen, tracking_metrics, writer
